# bodhi_update_state.py
#  - A utility class used to track the current state of an update
#    based on the history of its associated tests
#
# Copyright 2011, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#       Tim Flink <tflink@redhat.com>

import peak.util.symbols

from autoqa import util

# symbols used for consistant declaration of state
PASS = peak.util.symbols.Symbol('PASS', __name__)
FAIL = peak.util.symbols.Symbol('FAIL', __name__)
INCOMPLETE = peak.util.symbols.Symbol('INCOMPLETE', __name__)
UNKNOWN = peak.util.symbols.Symbol('UNKNOWN', __name__)

class BodhiUpdateState:

    def __init__(self, name, kojitag='updates'):
        self.test_states = {    'depcheck_32':'NORUN',
                                'depcheck_64':'NORUN',
                                'upgradepath':'NORUN'
                           }

        # since upgradepath is only run on updates-pending,
        # set it to passed if we're working on updates-testing
        if kojitag == 'updates-testing':
            self.test_states['upgradepath'] = 'PASSED'

        self.result_history = []
        self.update_name = name
        self.test_change = False

    def num_results(self):
        ''' Retrieve the number of results currently held '''
        return len(self.result_history)

    def add_result(self, new_result):
        ''' Update the current state with a new result.

            Args:
            new_result: a dictionary containing the following keys:
                time (datetime object), testname, result, arch
        '''

        self.result_history.append(new_result)

        # copy the current test state so that we can compare later
        old_state = self.test_states.copy()

        # for depcheck, we only care about i386 and x86_64 arches
        # anything else will be ignored
        if new_result['testname'] == 'depcheck':
            arch = util.get_basearch(new_result['arch'])

            if arch == 'i386':
                self.test_states['depcheck_32'] = new_result['result']

            elif arch == 'x86_64':
                self.test_states['depcheck_64'] = new_result['result']

            else:
                print "Arch %s is not valid for depcheck and will be ignored." \
                    % new_result['arch']

        # don't care about the arch for upgrade path. It should be noarch
        # but the results aren't changed if it isn't
        elif new_result['testname'] == 'upgradepath':
            self.test_states['upgradepath'] = new_result['result']

        # detect if any tests changed from the last result
        if old_state != self.test_states:
            self.test_change = True
        else:
            self.test_change = False

    def get_state(self):
        ''' Determine state based on currently known results '''

        # print out the current state for debugging purposes
        state_str = "Current State of update %s :" % self.update_name
        state_str += '\n  %s' % self.test_states
        for result in self.result_history:
            state_str += '\n  %s' % result
        print state_str + '\n'

        test_values = set(self.test_states.values())
        if 'NORUN' in test_values:
            return INCOMPLETE
        if 'FAILED' in test_values:
            return FAIL
        if 'PASSED' in test_values and len(test_values) == 1:
            return PASS

        return UNKNOWN

    def did_test_change(self):
        ''' Determine whether the state of any tests changed with the addition
            of the last result'''
        return self.test_change
