# Copyright 2009, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Will Woods <wwoods@redhat.com>

'''Global configuration for AutoQA and relevant helper methods.

All configuration is done statically, all options must be defined here in this
file in order to be accessible. They are not created dynamically from
configuration files contents.
'''

import os
import ConfigParser
import socket
import sys
import re

from autoqa import util, exc

# directory where all config files are stored
CONF_DIR = '/etc/autoqa'

# list of all AutoQA configuration files loaded into Config class
CONF_FILES = ['autoqa.conf',
              'secrets.conf',
             ]


class Config(object):
    '''Global configuration for AutoQA.

       All configuration options are available as class attributes in the form
       of Config.section.option = value. Different config files are not
       distinguished, all sections are flattened into a single name space.

       The documentation for individual options is available in their .conf
       file.

       Implementation notes:
       * All new options must be provided here, placing them just in a .conf
         file is not enough.
       * When changing defaults in here, don't forget to also update the
         corresponding .conf file.
       * Always assign a correct value type to an attribute (e.g. use empty
         string instead of None), because values read from .conf files are then
         automatically re-typed according to the attribute's current value type.
    '''

    class _Section(object):
        '''An empty class that represents a configuration section (relates to
        sections in .conf files). All config options are assigned as instance
        attributes.'''
        pass

    class _SectionPaths(_Section):
        '''A special version of @class _Section for configuring file paths. It
        makes sure that some selected paths exist after their assignment.'''

        def __init__(self):
            self._tmpdir = None

        @property
        def tmpdir(self):
            '''Temporary directory for generic AutoQA files (running tests have
            their own tmpdir provided by autotest). If you set a value, that
            directory is automatically created.'''
            return self._tmpdir

        @tmpdir.setter
        def tmpdir(self, value):
            self._tmpdir = value
            util.makedirs(self._tmpdir)


    ### config keys definition ###
    # == autoqa.conf ===
    general = _Section()
    general.local = False
    general.autotest_server = socket.gethostname()
    general.debug = False

    notifications = _Section()
    notifications.mail_from = ''
    notifications.smtpserver = 'localhost'
    notifications.result_email = ''
    notifications.result_email_blacklist = ''
    notifications.send_optin_email = False
    notifications.notification_email = ''
    notifications.send_bodhi_comments = False
    notifications.bodhi_posting_comments_span = 4320

    resources = _Section()
    resources.koji_url = 'http://koji.fedoraproject.org/kojihub'
    resources.pkg_url = 'http://kojipkgs.fedoraproject.org/packages'
    resources.bodhi_server = ''
    resources.git_baseurl = 'http://git.fedorahosted.org/git/'
    resources.http_proxy = ''
    resources.yourls_server = ''
    resources.resultsdb_server = ''

    bodhi_email = _Section()
    bodhi_email.email_all_passed = False
    bodhi_email.email_update_state_change = True
    bodhi_email.email_test_state_change = False
    bodhi_email.email_always_send = False
    bodhi_email.email_updates_testing = True

    paths = _SectionPaths()
    paths.eventdir = '/usr/share/autoqa/events'
    paths.watcherdir = '/usr/share/autoqa/watchers'
    paths.testdir = '/usr/share/autotest/client/site_tests'
    paths.cachedir = '/var/cache/autoqa'
    paths.tmpdir = '/tmp/autoqa'
    paths.lockdir = '/var/run/autoqa'
    paths.logfile = '/var/log/autoqa/autoqa.log'
    paths.repoinfo_conf = os.path.join(CONF_DIR, 'repoinfo.conf')
    paths.log_template = os.path.join(CONF_DIR, 'log_template.html')

    # == secrets.conf ==
    secrets = _Section()
    secrets.fas_username = 'autoqa'
    secrets.fas_password = ''
    secrets.yourls_username = ''
    secrets.yourls_password = ''


    @classmethod
    def _load(cls, filepath):
        '''Load options from a configuration file and then assign their values
           to attributes in this class.

           Empty values are ignored. Extra options and sections (not existing
           here) are also ignored.

           @param filepath conf file path (type: string)
        '''
        hint = ('Maybe you have an outdated configuration file and want to '
                'merge it with the latest version? Or if you want to define a '
                'new option, you have to edit autoqa.config module as well.')

        parser = ConfigParser.SafeConfigParser()
        parser.read(filepath)

        for section in parser.sections():
            if not hasattr(cls, section):
                print >> sys.stderr, ('WARN: Unknown section "%s" in config '
                                      'file "%s"' % (section, filepath))
                print >> sys.stderr, hint
                continue
            section_conf = getattr(cls, section)

            options = parser.items(section)
            for (option, value) in options:
                if not hasattr(section_conf, option):
                    print >> sys.stderr, ('WARN: Unknown option "%s" in section '
                        '"%s" in config file "%s"' % (option, section, filepath))
                    print >> sys.stderr, hint
                    continue

                # Only set option if it has a non-empty value.
                # That allows users to leave 'option = ' lines in config file and
                # still use default values. OTOH user can't empty out a default
                # option value this way.
                if value:
                    # re-type to correct type
                    try:
                        value = cls._retype(getattr(section_conf, option), value)
                    except exc.AutoQAConfigError, e:
                        print >> sys.stderr, hint
                        raise exc.AutoQAConfigError('Incorrect value provided '
                            'for option "%s"' % option, e)

                    # assign value
                    setattr(section_conf, option, value)


    @classmethod
    def _retype(cls, old_value, new_value):
        '''Re-type new_value according to the current type of old_value.

           @return re-typed new_value (it may be the same if old_value was of
                   the same type or None); if exception occurs during
                   conversion then None is returned
        '''

        if old_value is not None:
            try:
                type_ = type(old_value)
                # for booleans do advanced checking
                if issubclass(type_, bool):
                    newvalue = util.getbool(new_value)
                else:
                    newvalue = type_(new_value)
                return newvalue
            except StandardError, e:
                raise exc.AutoQAConfigError("Can't re-type '%s' to %s" %
                                            (new_value, type_), e)

        return new_value


    @classmethod
    def init(cls):
        '''Initialize the class and load values from all .conf files. This
        method must be called before accessing class attributes. The call
        should be ideally placed into package initialization module, as soon as
        possible.
        '''
        # load configuration files
        for conffile in CONF_FILES:
            cls._load(os.path.join(CONF_DIR, conffile))



def use_proxy(use=True):
    '''Enable or disable HTTP proxy as provided in the config file. This is
       done by setting http_proxy environment variable.
       @param use True or False; if True is given but no proxy is provided
                  in the conf file, then nothing happens
    '''
    proxy = Config.resources.http_proxy
    if use:
        if proxy:
            # definition of hosts that should be never proxied
            no_proxy = set(['localhost', '127.0.0.1'])

            no_proxy.add(socket.gethostname())
            try:
                no_proxy.add(socket.gethostbyname(socket.gethostname()))
            except socket.error:
                pass

            no_proxy.add(Config.general.autotest_server)
            try:
                no_proxy.add(socket.gethostbyname(Config.general.autotest_server))
            except socket.error:
                pass

            no_proxy = no_proxy.difference(set([None, '']))
            no_proxy = ','.join(no_proxy)

            # set variables
            os.environ['http_proxy'] = proxy
            os.environ['no_proxy'] = no_proxy
    else:
        os.environ.pop('http_proxy', None)
        os.environ.pop('no_proxy', None)


def email_blacklisted(subject):
    '''Check whether provided email is blacklisted or not in AutoQA configuration.
       @param subject email subject
       @returns True if email is blacklisted and should not be sent, False
                otherwise
    '''
    patterns = Config.notifications.result_email_blacklist
    patterns = patterns.splitlines()

    # go through all the patterns and test them against email subject
    for pattern in patterns:
        pattern = pattern.strip()
        if not pattern:
            continue
        if re.match(pattern, subject):
            return True

    return False
