import sys
import traceback

class ExceptionCatcher(object):
    """
    Because autotest-driven tests immediately end, when exception
    is raised, use this as a decorator around the run_once function.
    If an exception is raised in the decorated function, this decorator
    captures it, then runs the function specified by on_exception, and
    finally re-raises the exception.

    @ExceptionCatcher(on_exception, *args, **kwargs)

    on_exception parameter can be either pointer to function or string
    if it's a string, it's evaluated as if it was a name of a class method.

    Example:

        >>> def helper_func():
        >>>     print "Outside Foo!"

        >>> class Foo(object):
        >>>     @ExceptionCatcher("self.helper_func") #or "helper_func"
        >>>     def some_func_1(self):
        >>>         #calls self.helper_func on exception
        >>>         raise Exception()

        >>>     @ExceptionCatcher(helper_func) #notice the missing quotes
        >>>     def some_func_2(self):
        >>>         #calls the  outside helper_func on exception
        >>>         raise Exception()

        >>>     def helper_func(self):
        >>>         print "Inside Foo!"

        >>> x = Foo()
        >>> x.some_func_1()
        Inside Foo!
        Traceback (most recent call last):
        ...
        Exception

        >>> x.some_func_2()
        Outside Foo!
        Traceback (most recent call last):
        ...
        Exception
    """

    def __init__(self, on_exception = "self.process_exception", *args, **kwargs):
        """
        on_exception - string with class function name, or pointer to function
        *args, **kwargs - args for the on_exception function
        """
        self.oe = on_exception
        #if the on_exception is string, then it's handled as a
        #class member function
        self.oe_is_self_method = type(self.oe) is type("")
        if self.oe_is_self_method and self.oe.startswith('self.'):
            self.oe = self.oe[5:]
        self.oe_a = args
        self.oe_k = kwargs

    def __call__(self, f):
        def newf(*args, **kwargs):
            #if called with arguments, first is stored as
            #'self' reference, in case it's needed
            if len(args) > 0:
                f_self = args[0]
            exc_info = None
            try:
                f_result = f(*args, **kwargs) #call the decorated function
            except:
                exc_info = sys.exc_info() #store the exception
            finally:
                if exc_info is not None: #if exception happened
                    try:
                        #if the 'on_exception' is a name of 'class member function'
                        #then call the f_self.on_exception(oe_args, oe_kwargs)
                        if self.oe_is_self_method:
                            f_self.__getattribute__(self.oe)(exc = exc_info, *self.oe_a, **self.oe_k)
                        #'on_exception' is pointer to function
                        else:
                            self.oe(exc = exc_info, *self.oe_a, **self.oe_k)
                    except Exception:
                        traceback.print_exc(file=sys.stderr)
                    finally:
                        #re-raise the stored exception
                        raise exc_info[0], exc_info[1], exc_info[2]
                return f_result
        return newf

