#!/usr/bin/python

# Copyright 2011, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Josef Skladanka <jskladan@redhat.com>

import os
import fcntl

class FileLock:
    """Simple API for locking files using fcntl"""
    def __init__(self, filename):
        """@param filename Name of the file to be locked"""
        self.filename = filename
        self._fd = None
        self.lock_acquired = False

    def acquire(self):
        """Tries to acquire lock on the file. If the file
        specified in self.filename does not exist, it creates the file.
        If this process dies (or is killed), the lock is released.
        (yes, even if SIGKILLed)

        @returns True if lock was successfully acquired
        @returns False if lock was not acquired"""
        if self.lock_acquired:
            # lock already acquired
            return True

        if not os.path.isfile(self.filename):
            try:
                self._fd = os.open(self.filename, os.O_CREAT|os.O_EXCL|os.O_RDWR)
            except OSError:
                return False
        else:
            self._fd = os.open(self.filename, os.O_RDWR)

        try:
            fcntl.flock(self._fd, fcntl.LOCK_EX|fcntl.LOCK_NB)
            self.lock_acquired = True
            return True
        except IOError:
            self._fd = None
            return False


    def release(self):
        """Releases the lock
        @returns True if lock was successfully released
        @returns False otherwise"""
        if not self._fd:
            return False
        try:
            os.close(self._fd)
            self._fd = None
            self.lock_acquired = False
            return True
        except OSError:
            return False

    def __del__(self):
        self.release()

def main():
    lock = FileLock("/tmp/lock.file")
    while 1:
        cmd = raw_input("(a)cquire lock, (r)elease lock: ").strip()
        if cmd == 'a':
            if lock.acquire():
                print "Lock acquired"
            else:
                print "Unable to acquire lock"
        elif cmd == 'r':
            if lock.release():
                print "Lock released"
            else:
                print "Unable to release lock"

if __name__ == "__main__":
    main()

