#!/usr/bin/python
#
# Copyright 2011, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Martin Krizek <mkrizek@redhat.com>

'''This is a simple module providing a logger to autoqa script and watchers.'''

import sys
import traceback
import logging
import optparse

from autoqa.config import Config


def _log_excepthook(*exc_info):
    '''Called when an exception is not caught'''
    logging.critical(''.join(traceback.format_exception(*exc_info)))

# overwrite default except hook
sys.excepthook = _log_excepthook


def _get_level(opts):
    '''
    Get log level according to the cmd line options.

    Args:
    opts - optparse.Values object
    '''
    loglevel = logging.INFO # verbose is default
    if opts.debug:
        loglevel = logging.DEBUG
    elif opts.quiet:
        loglevel = logging.ERROR
    return loglevel


def add_options(parser):
    '''
    Add command line options controlling log level.

    Args:
    parser - optparse.OptionParser instance
    '''
    log = optparse.OptionGroup(parser, "Verbose options")
    log.add_option('-d', '--debug', action="store_true",
        help='Show debug messages')
    log.add_option('-q', '--quiet', action="store_true",
        help='Suppress messages except errors')
    parser.add_option_group(log)


def setup_logger(level=logging.INFO, file_logger=True, opts=None):
    ''' Return a logger instance.

        Args:
        level - minimum log level to be printed, provide
                one of @attr logging.{DEBUG,INFO,WARN,ERROR,CRITICAL}
        file_logger - indicates whether write into the log file or not
        opts - optparse.Values object. If provided, the chosen log level
               will be parsed from it and @param level will be ignored.
               See also @method add_options.
    '''

    if opts is not None:
        level = _get_level(opts)

    root = logging.getLogger()
    root.setLevel(logging.NOTSET) # clear the existing value

    # Handler for standard output
    # StreamHandler uses default formatter - just plain message
    sh = logging.StreamHandler()
    sh.setLevel(level)
    root.addHandler(sh)

    if not file_logger:
        return

    # Handler for a log file
    # Everything is logged in the log file except debug msgs
    # unless debug option is set in autoqa.conf, then also
    # debug msgs are logged.
    file_lvl = logging.DEBUG if Config.general.debug else logging.INFO
    logfile = Config.paths.logfile

    try:
        fh = logging.FileHandler(logfile)
    except IOError, e:
        logging.warning("%s", e)
        return
    fh.setLevel(file_lvl)
    log_format = '[%(filename)s:%(lineno)d] ' if level == logging.DEBUG else '%(module)-12s '
    log_format += '%(asctime)s %(levelname)-7s %(message)s'
    formatter = logging.Formatter(fmt=log_format, datefmt='%Y-%m-%d %H:%M:%S')
    fh.setFormatter(formatter)
    root.addHandler(fh)
