#!/usr/bin/python

# Copyright 2009, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Vitezslav Humpa <vhumpa@redhat.com>

import sys

from time import gmtime, strftime
from string import Template
from cgi import escape

import test   # for some reason 'from autoqa import test' doesn't work here
from autoqa.config import Config


class PrettyLog(object):
    """
    This class keeps and generates "pretty" logs. Pretty logs should be easily
    readable, as opposed to a plain test output.

    @attr test_detail TestDetail instance to access basic information at:
                      id, result, summary, arch, output
    @attr summary_detail the body of the summary section; either a string or a
                         dictionary of results {result: [items]} where keys are
                         either result keywords (default result order is used
                         for sorting) or arbitrary strings (then alphabetical
                         order is used). 'items' are strings (e.g. package
                         builds). If a dictionary is provided
                         and test_detail.summary is empty, the summary
                         is generated out of it and set.
    @attr id_name the name of the "ID" field in the test overview
    @attr kojitag "Koji tag" field in the test overview
    @attr bodhi_title "Bodhi update" field in the test overview
    @attr bodhi_id "Bodhi ID" field in the test overview
    @attr custom_fields a dictionary of custom fields for the test overview
                        section. Dictionary key is field name and dictionary
                        value is field value.
    """

    def __init__(self, test_detail, summary_detail=None,
                 id_name="ID", kojitag=None, bodhi_title=None, bodhi_id=None,
                 custom_fields=None):
        """Create new PrettyLog.
        For parameter description see class docstring.
        """

        self.test_detail = test_detail
        self._summary_detail = summary_detail
        self.id_name = id_name
        self.kojitag = kojitag
        self.bodhi_title = bodhi_title
        self.bodhi_id = bodhi_id
        self.custom_fields = custom_fields

        # load the pretty log template
        with open(Config.paths.log_template, 'r') as file_:
            self._template = file_.read()


    def _make_summary(self):
        '''Construct the summary header for multi-item tests'''

        results = []
        for result, array in self.summary_detail.items():
            results.extend([result] * len(array))
        return test.create_multi_item_summary(results)


    def _make_summary_detail(self):
        '''Construct the body of the summary section for multi-item tests'''

        detail = ''
        # if all keys are result keywords then they are sorted according
        # to their priority. Otherwise the keys are sorted alphabetically.
        td = test.TestDetail(test = None, id = None)
        results = sorted(self.summary_detail.keys(),
                         cmp = td.cmp_result
                               if set(self.summary_detail.keys()).issubset(set(td.result_order))
                               else cmp
                         )

        for result in results:
            if not self.summary_detail[result]:
                continue
            detail += '%s:\n' % result
            for item in self.summary_detail[result]:
                detail += '  %s\n' % item
        return detail.rstrip()


    def _make_highlights(self, output, id_=1):
        '''Constructs the highlights section and modifies output to highlight
           selected text.
           @param output - an output string
           @returns a highlight section string
        '''

        def highlight_line(id_, line, comment, output):
            '''Highlights a line in the Detailed results section
               and adds an optional comment right to the line
            '''

            comment_exp = '<span class="comment">&gt;&gt; %(comment)s' \
                        '</span>' % {'comment': comment} if comment else ''
            new_output = '\n%s\n' % output
            # we allow only for a complete line(s)
            new_output = new_output.replace('\n%s\n' % line,
                    '\n<span id="%(id)s" class="highlight">%(line)s' \
                    '</span> %(comment)s\n' % {'id': id_,
                                               'line': line,
                                               'comment' : comment_exp})
            return new_output.strip('\n'), new_output != output

        #heading of the section
        highlights = '<h3>Highlights</h3>\n<pre id="padded">\n'
        #template of the line in highlights section
        highlights_line = '{} <a href="#{}">[view &raquo;]</a>\n'
        processed = []
        for (line, comment) in self.test_detail.highlights:
            # we don't want to change self.test_detail, that's why escape is here
            if line:
                line = escape(line)
            if comment:
                comment = escape(comment)
            if line in processed:
                pass
            elif line:
                #highlight a line in output
                output, highlid = highlight_line(id_, line, comment, output)
                #add a line into the highlights section, if found previously
                if highlid:
                    highlights += highlights_line.format(line, id_)
                else:
                    print >> sys.stderr, "Can't highlight, no such line: %s" % line
                #id for infile hyperlinking
                id_ += 1
                processed.append(line)
            elif comment:
                #just a comment present > append to highlights section
                highlights += comment + '\n'

        return output, highlights + '</pre>'


    def _make_overview(self, plain_text=False, summary=False):
        '''Construct the overview section (the top of the log)'''

        fields = []

        # ID
        fields.append((self.id_name, self.test_detail.id))
        # Result
        fields.append(('Result', self.test_detail.result))
        # Summary
        if summary:
            fields.append(('Summary', self.test_detail.summary))
        # Koji tag
        if self.kojitag:
            fields.append(('Koji tag', self.kojitag))
        # Bodhi update title
        if self.bodhi_title:
            fields.append(('Bodhi update', self.bodhi_title))
        # Bodhi update ID
        if self.bodhi_id:
            fields.append(('Bodhi update ID', self.bodhi_id))
        # Architecture
        arch = self.test_detail.arch
        if arch == 'noarch':
            arch = 'all'
        fields.append(('Architecture', arch))
        # Custom fields
        if self.custom_fields:
            for key, value in self.custom_fields.items():
                fields.append((key, value))
        # Test ended time
        fields.append(('Test ended', strftime("%Y-%m-%d %H:%M:%S UTC", gmtime())))
        # Full log link: html or plain
        fields.append(('Full log', self.test_detail.test.url['full_log']))

        # Escape html related characters and control for None
        fields = [
                    [None if x is None else escape(x),
                     None if y is None else escape(y)]
                    for x, y in fields
                 ]

        # Now that all text is escaped, we can post-process some fields and add
        # HTML tags to some of them
        if not plain_text:
            # colorize result
            field = [field for field in fields if field[0] == 'Result'][0]
            field[1] = '<span style="color: %s">%s</span>' % (
                self._result_color(field[1]), field[1])
            # change full log URL to hyperlink
            field = [field for field in fields if field[0] == 'Full log'][0]
            field[1] = self._wrap_link(field[1])

        # generate the final string
        overview = ''
        field = " {:<20}{}\n" if plain_text else \
            '<tr>\n<th align="left">{}</th>\n<td id="padded">{}</td>\n</tr>\n'
        for key, value in fields:
            overview += field.format(key, value)

        return overview.rstrip()


    def _wrap_link(self, hyperlink):
        '''Wrap string hyperlink with HTML <a> tag.'''

        return '<a href="%(link)s">%(link)s</a>' % {'link': hyperlink}


    def _result_color(self, result):
        '''Return an HTML-style color (as a string) that matches provided result
        keyword (string).'''

        colors = {
            'PASSED': 'green',
            'INFO': 'olive',
            'FAILED': 'red',
            'NEEDS_INSPECTION': 'red',
            'ABORTED': 'magenta',
            'CRASHED': 'magenta',
        }

        if result in colors:
            return colors[result]
        else:
            return 'black'


    def plain_overview(self, include_summary=False):
        '''Returns a simple plain-text overview of the log.'''

        return '%s\n AutoQA %s Test Results\n%s\n%s' % (
                   '=' * 72,
                   self.test_detail.test.__class__.__name__.title(),
                   '-' * 72,
                   self._make_overview(plain_text=True,summary=include_summary))


    def create(self):
        """
        Return the final pretty log string.
        """
        template = Template(self._template)

        # convert summary_detail is necessary
        summary_detail = self.summary_detail or ''
        if isinstance(self.summary_detail, dict):
            summary_detail = self._make_summary_detail()

        output = '\n'.join(self.test_detail.output)
        output = escape(output)
        highlights = ''

        if self.test_detail.highlights:
            output, highlights = self._make_highlights(output)

        # escape html characters in all remaining strings before creating
        # the final document
        content = dict(test = escape(self.test_detail.test.__class__.__name__.title()),
                       overview = self._make_overview(),
                       summary = escape(self.test_detail.summary or ''),
                       summary_detail = escape(summary_detail),
                       highlights = highlights,
                       result_detail = output)

        pretty_log = template.substitute(content)
        return pretty_log


    @property
    def summary_detail(self):
        return self._summary_detail


    @summary_detail.setter
    def summary_detail(self, value):
        self._summary_detail = value
        # if summary is empty and user provided dict of summary_detail, let's
        # create summary
        if not self.test_detail.summary and isinstance(self.summary_detail, dict):
            self.test_detail.summary = self._make_summary()
