#
# repoinfo.py - get repository info from repoinfo.cfg for autotest watchers
#
# Copyright 2009, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Will Woods <wwoods@redhat.com>

import ConfigParser
import warnings

from autoqa.config import Config

class RepoinfoConfig(object):
    '''This class is a wrapper for easily accessing repoinfo.conf file.'''

    def __init__(self, arch="$basearch", filelist=Config.paths.repoinfo_conf,
                 defaults=None):
        '''Create RepoinfoConfig object.
        Attributes:
        * arch - architecture for which to adjust repo URLs
        * filelist - list of config files to read information from
        * defaults - dictionary of default values
        '''
        if defaults is None:
            self.defaults = {}
        else:
            self.defaults = defaults
        self.filelist = filelist
        self.setarch(arch) # implies read()


    def _getlist(self, name, item):
        '''Transform a list of comma separated values into a Python list
        Returns empty list for non-existent item'''
        itemlist = self.config.get(name, item)
        if not itemlist:
            return []
        else:
            return [t.strip() for t in itemlist.split(',')]


    def getarches(self, name):
        '''Get the list of all supported architectures for a repo'''
        return self._getlist(name, 'arches')


    def getreleases(self):
        '''Return the list of known, non-EOL releases.'''
        return [r for r in repoinfo.repos() if self.config.get(r,'isactiverelease')]


    def get_bodhi_releases(self):
        '''Return the list of releases known to Bodhi.'''
        return [r.upper() for r in self.getreleases()]


    def getparents(self, name):
        '''Given a repo name, give the names of its parent repos
        e.g.: getparents('f13-updates-testing') -> ['f13','f13-updates']'''
        return self._getlist(name, 'parents')


    def get_top_parents(self, reponame):
        '''Get top parents reponames for a specified reponame.
        @param reponame repository name
        @return list of top parent reponames (list of strings). Usually this
            is a single-item list, but in some cases there may be more
            than one top parent.
        '''
        parents = self.getrepo(reponame)['parents']
        if not parents:
            return set([reponame])
        else:
            result = set()
            for reponame in parents:
                result.update(self.get_top_parents(reponame))
            return result


    def getrepo(self, name):
        '''Given a repo name, return the repoinfo dict, with the keys:
        'arches', 'parents', 'tag', 'url', 'path', 'name', 'collection_name'
        and 'composes' '''
        repo = {'arches': self.getarches(name),
                'parents': self.getparents(name),
                'tag': self.config.get(name,'tag'),
                'url': self.config.get(name,'url'),
                'path': self.config.get(name,'path'),
                'name': name,
                'collection_name': self.config.get(name, 'collection_name'),
                'composes': self.config.getboolean(name,'composes'),
        }
        return repo


    def getrepo_by_tag(self, tag):
        '''Given a koji tag, return the corresponding repo
        e.g.: getrepo_by_tag('dist-f13')'''
        # we don't have repo definition for -candidate tags,
        # let's go up in the hierarchy (stripping '-candidate')
        if tag.endswith('-candidate'):
            tag = tag[:-len('-candidate')]
        for r in self.repos():
            if self.config.get(r,'tag') == tag:
                return self.getrepo(r)


    def repos(self):
        '''Returns a list of all known repo names'''
        return self.config.sections()


    def setarch(self, arch):
        '''Set architecture for which to adjust repo URLs'''
        self.defaults['arch'] = arch
        self.read()


    def read(self):
        '''Apply the defaults and then read first available config file from
        the list of provided config files.'''
        self.config = ConfigParser.SafeConfigParser(defaults=self.defaults)
        self.config.read(self.filelist)


    def get(self, repo, key):
        '''Get a specific key from a repo'''
        return self.config.get(repo, key)


repoinfo = RepoinfoConfig()

if not repoinfo.repos():
    warnings.warn("No repos found in (%s)" % ', '.join(Config.paths.repoinfo_conf),
                  RuntimeWarning)

if __name__ == '__main__':
    import pprint
    rd = dict([(n, repoinfo.getrepo(n)) for n in repoinfo.repos()])
    pprint.pprint(rd)
