# -*- coding: utf-8 -*-
#
# Copyright 2010, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Josef Skladanka <jskladan@redhat.com>

'''This module contains the main AutoQATest class from which all tests are
derived, and also some other classes and methods that are directly connected
to tis class.'''

import os
import sys
import traceback
import collections
import urllib
import stat
import yourls.client

from autotest_lib.client.bin import test, utils
from resultsdb_api import ResultsDBapi
from yourls import YourlsError, YourlsOperationError

from autoqa import bodhi_utils, util, exc, config
from autoqa.config import Config
from autoqa.decorators import ExceptionCatcher
from autoqa.repoinfo import repoinfo
from autoqa.pretty_log import PrettyLog


class AutoQATest(test.test, object):
    '''The main class from all tests are derived.

    @attr detail instance of TestDetail. This is the main instance where all
                 test outcomes are stored at.
    @attr config dictionary with a few important run-time properties, which
                 were passed by the server
    @attr url dictionary of URLs denoting the locations where results of this
              test will be accessible:
              { 'base': base URL to the main autotest directory with all
                        information (string)
                'debug_dir': URL to autotest debug/ directory (string)
                'result_dir': URL to autotest results/ directory (string)
                'full_log': URL to test's full.log file (string)
              }
              All directories end with a slash.
    @attr is_multiresult set to True in order to tell the AutoQA to handle the
                         result reporting in a slightly different way, which
                         is required for the multi-result tests like depcheck
    '''

    version = 1 # increment if setup() changes
    testcase_url = "https://fedoraproject.org/wiki/User:Jskladan/Sandbox:EMPTY_Testcase_Metadata"


    def __init__(self, *args, **kwargs):
        super(AutoQATest, self).__init__(*args, **kwargs)

        self.detail = TestDetail(test=self, id=None)
        # Denotes whether any TestDetail results were already posted. Important
        # when deciding whether to create PrettyLog out of self.detail or not.
        self._any_detail_posted = False
        self.is_multiresult = False

    @ExceptionCatcher()
    def setup(self, **kwargs):
        pass


    @ExceptionCatcher()
    def initialize(self, config, **kwargs):
        '''Initializes some internal variables and creates new entry in ResultsDB
        via 'start_testrun()' method. 'end_testrun()' is called during
        'post_results()' phase.'''
        self.config = config
        self.url = self._create_url_dict()
        self._yourls_client = self._get_yourls_client()

        self._use_resultsdb = Config.resources.resultsdb_server != ""
        if self._use_resultsdb:
            try:
                self._resultsdb_client = ResultsDBapi(Config.resources.resultsdb_server)
                # stores the resultsdb's job id, it's used to link testruns to the job
                self._job_id = None
                if self.is_multiresult:
                    self._job_id = self._resultsdb_client.start_job(name = self.__class__.__name__,
                                                             log_url = self.url['debug_dir'],
                                                            )
                else:
                    self.detail._resultsdb['testrun_id'] = self._resultsdb_client.start_testrun(
                                                               testcase_url = self.testcase_url,
                                                               keyval_pairs = {"testname": self.__class__.__name__},
                                                               job_id = self._job_id,
                                                               log_url = self.url['debug_dir'],
                                                               )
            except:
                print >> sys.stderr, "ResultsDB initialization failed;"
                print >> sys.stderr, traceback.print_exception(*sys.exc_info())
                self._use_resultsdb = False


    def _create_url_dict(self):
        '''Create dictionary of URLs that is to be assigned to @attr url.'''
        base = "http://%s/results/%s/" % (Config.general.autotest_server,
                                          self.config['jobtag'])
        debug_dir = base + 'debug/'
        result_dir = '%s%s/results/' % (base, self.__class__.__name__)
        full_log = result_dir + 'full.log'
        url = dict(base=base, debug_dir=debug_dir, result_dir=result_dir,
                    full_log=full_log)
        return url


    def _get_yourls_client(self):
        '''Create and return an instance of YourlsClient, or None if some
           error occurs'''

        if not Config.resources.yourls_server:
            return None

        try:
            client = yourls.client.YourlsClient(Config.resources.yourls_server,
                Config.secrets.yourls_username, Config.secrets.yourls_password)
            return client
        except YourlsError as error:
            print >> sys.stderr, 'Error creating YOURLS client: %s' % error
            print >> sys.stderr, 'URLs will NOT be shortened'

        return None


    @ExceptionCatcher()
    def run_once(self, **kwargs):
        os.chdir(self.bindir) # easiest way for tests to find their test scripts,
                              # config files, etc


    def process_exception(self, exc):
        '''Called when an uncaught exception occurs in any method decorated
        with @ExceptionCatcher'''

        # limit the length of the exception string when putting into summary
        lines = str(exc[1]).splitlines()
        summary = lines[0] if lines else '(No explanation)'
        if len(summary) > 200:
            summary = summary[:100] + '...' + summary[-100:]

        if self.detail.result == 'ABORTED':
            # the exception is raised intentionally, don't override anything,
            # only if empty
            if not self.detail.summary:
                self.detail.summary = '%s: %s' % (exc[1].__class__.__name__,
                                                  summary)
        else:
            self.detail.result = 'CRASHED'
            self.detail.summary = '%s: %s' % (exc[1].__class__.__name__, summary)

        # append traceback
        msg = 80 * '-' + '\n'
        msg += ''.join(traceback.format_exception(exc[0], exc[1], exc[2]))
        self.log(msg, stderr=True, highlight=True)

        # do the rest as usual
        self.postprocess()


    def postprocess(self):
        '''Overridden method from Autotest that will finish up test execution
        (mainly call @method post_results()).'''

        self.log('%s %s %s' % (20*'*', 'TEST ENDED', 20*'*'), output=False)
        print self.detail.create_summary_table()

        # create log from self.detail only if not already some log was posted
        # manually
        if not self._any_detail_posted:
            self.post_results()

        # create full.log
        self._post_full_log()

        # ensure all logs have read permissions
        self._make_results_readable()


    def _make_results_readable(self):
        '''Ensure that all directories and files in the results dir are readable
        for all users.
        '''
        for root, dirs, files in os.walk(self.resultsdir):
            for d in (dirs):
                dirpath = os.path.join(root, d)
                os.chmod(dirpath, stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH |
                                  stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)
            for f in (files):
                filepath = os.path.join(root, f)
                os.chmod(filepath, stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)


    def _shorten_url(self, url, test_detail):
        '''Return shortened version of given url on success,
        otherwise return given url.

        @param url url to be shortened
        @param test_detail TestDetail instance
        '''

        if self._yourls_client is None:
            return url

        try:
            url = self._yourls_client.shorten(url)
        except YourlsOperationError as error:
            print >> sys.stderr, "Error shortening URL: %s" % error

        return url


    def post_results(self, test_detail=None, bodhi=None, optin=None):
        '''Send test results to various places.

        By default these actions are executed:
          'log' -- create log with test results
          'email' -- send email(s) with test results
          'resultsdb' -- send report to resultsdb
        Additional actions may be triggered, if you provide corresponding parameters:
          'bodhi' -- send comment to Bodhi (if enabled)
          'optin' -- send emails also to opted-in maintainers (if enabled and
                     applicable)

        @param test_detail TestDetail for which to submit results. If None, then
                           it uses the main TestDetail (self.detail)
        @param bodhi if you want to send a comment to Bodhi, you must provide a
                     dictionary:
                     {
                       'title' : title of the Bodhi update to comment on (string)
                     }
        @param optin if you want to send emails also to opted-in maintainers,
                     you must provide a dictionary:
                     {
                       'pkg': name of the package tested; the result email will
                              be sent to its owner (string)
                       'kojitag': Koji tag related to the pkg; email will be sent
                                  only if maintainer opted-in for matching release
                                  (string)
                     }
        '''

        if test_detail is None:
            test_detail = self.detail

        # which actions to take?
        dest = ['log', 'email']
        if bodhi:
            dest.append('bodhi')
        if optin:
            dest.append('optin')
        if self._use_resultsdb:
            dest.append('resultsdb')

        self.log('Submitting results (%s) for: %s' % (', '.join(dest),
                 test_detail.excerpt()), output=False)

        # write log
        logname = self._write_log(test_detail)

        # send email (optin is optional)
        self._send_email(test_detail, optin=optin, logname=logname)

        # send to ResultsDB
        log_url = self.url['result_dir'] + logname
        if self._use_resultsdb:
            self._send_resultsdb(test_detail, log_url)

        # more optional stuff
        if bodhi:
            log_url = self._shorten_url(self.url['result_dir'] + logname, test_detail)
            self._send_bodhi(test_detail, bodhi['title'], log_url)

        # mark details as posted
        self._any_detail_posted = True


    def _post_full_log(self):
        dest = ['log', 'email']
        self.log('Submitting full log (%s) for: %s' % (', '.join(dest),
                 self.detail.excerpt()))

        # write log
        logname = self._write_log(self.detail, full_log=True)

        log_url = self.url['result_dir'] + logname
        if self._use_resultsdb:
            try:
                # end the job in case of multiresult test
                if self.is_multiresult:
                    self._resultsdb_client.end_job(self._job_id, result = self.detail.result, log_url = log_url, summary = self.detail.summary, sync_testrun_times = True)
            except:
                print >> sys.stderr, "ResultsDB end_job failed; %s" % self._job_id
                print >> sys.stderr, traceback.print_exception(*sys.exc_info())

        # send email
        self._send_email(self.detail, full_log=True)


    def _send_bodhi(self, test_detail, title, log_url):
        '''Send results as a Bodhi comment.

        @param test_detail TestDetail of the test run
        @param title title of the Bodhi update
        @param log_url hyperlink where the log file is located
        '''
        bodhi_utils.bodhi_post_testresult(title, self.__class__.__name__,
            test_detail.result, log_url, test_detail.arch, karma=0)

    def _send_resultsdb(self, test_detail, log_url):
        ''' Send results to ResultsDB via 'end_testrun()' method.
        If the test_detail.resultsdb['testrun_id'] is not set,
        then 'start_testrun()' is called in advance.

        @param test_detail TestDetail of the test run
        @param log_url hyperlink where the log file is located
        '''

        # If the test_detail has no associated testrun
        # i.e. it's a test_detail created by multi-result tests
        # so we need to start the testrun here
        # start & end time of the testrun is synchronized
        # from the end_job() call in _post_full_log()
        if test_detail._resultsdb['testrun_id'] in [0, None]:
            try:
                test_detail._resultsdb['testrun_id'] = self._resultsdb_client.start_testrun(
                                                       testcase_url = self.testcase_url,
                                                       keyval_pairs = {
                                                                        "testname": self.__class__.__name__,
                                                                      },
                                                       job_id = self._job_id,
                                                       )
            except:
                print >> sys.stderr, "ResultsDB _send_resultsdb() start_testrun failed"
                print >> sys.stderr, traceback.print_exception(*sys.exc_info())
                return None

        kp = {"arch": test_detail.arch}
        kp.update(test_detail.keyvals)
        try:
            self._resultsdb_client.end_testrun(test_detail._resultsdb['testrun_id'], test_detail.result, log_url, kp, test_detail.summary)
        except:
            print >> sys.stderr, "ResultsDB _send_resultsdb() end_testrun failed; %s" % test_detail._resultsdb['testrun_id']
            print >> sys.stderr, traceback.print_exception(*sys.exc_info())

    def _send_email(self, test_detail, optin=None, full_log=False, logname=None):
        '''Send results via email.

        @param test_detail TestDetail of the test run
        @param optin the same param as in @method post_results(); if provided
                     and opt-in emails are enabled and applicable, they will
                     be also sent
        @param full_log whether to send full.log from the specified TestDetail
                        or a standard log
        @param logname file name of the (Pretty) log. Only needed and used when
                       @param full_log=False. The log name is used for
                       constructing URL and putting it into the email body.
        '''
        # collect text
        mail_body = self._create_email_body(test_detail, full_log=full_log,
                                            logname=logname)
        subject = test_detail.excerpt()

        # collect recipients
        send_to = list(test_detail.email_to)
        result_email = Config.notifications.result_email
        if result_email and full_log and not config.email_blacklisted(subject):
            send_to.extend(result_email.split(','))
        if optin:
            repo = repoinfo.getrepo_by_tag(optin['kojitag'])
            if (repo is not None and Config.notifications.send_optin_email and
                util.check_opt_in(optin['pkg'], repo['collection_name'])):
                send_to.append('%s-owner@fedoraproject.org' % optin['pkg'])

        # send email
        if send_to:
            utils.send_email(mail_to=send_to,
                mail_from=Config.notifications.mail_from,
                mail_server=Config.notifications.smtpserver,
                subject=subject,
                body = mail_body)
            print 'Email sent: "%s" to %s' % (subject, ','.join(send_to))
        else:
            print 'Email not sent: "%s"' % subject


    def _create_email_body(self, test_detail, full_log=False, logname=None):
        '''Create output suitable for email.

        @param test_detail TestDetail of the test run
        @param full_log whether to send full.log from the specified TestDetail
                        or a standard log
        @param logname file name of the (Pretty) log. Only needed and used when
                       @param full_log=False. The log name is used for
                       constructing URL and putting it into the email body.
        '''
        # validate input
        if not full_log and not logname:
            raise ValueError('logname parameter required when full_log=False')

        if full_log:
            # the email body is the same as the full.log
            return test_detail.create_output(full_log=True)
        else:
            # we don't want to send HTML log inside email. Instead let's create
            # a plaintext header and put a link to the HTML log inside.
            overview = test_detail.log.plain_overview(include_summary=True)
            # we don't want to have the link to full.log in there, that would be
            # very confusing (in addition to HTML log link). Let's strip it.
            overview = [line for line in overview.splitlines() if
                        self.url['full_log'] not in line]
            # now add the link to the HTML log
            overview.append('')
            url = self._shorten_url(self.url['result_dir'] + urllib.quote(logname), test_detail)
            overview.append('This is just an overview extract of the test '
                'results, please see the log for more details:\n%s'
                % url)
            overview = '\n'.join(overview)
            return overview


    def _write_log(self, test_detail, full_log=False):
        '''Create log for a TestDetail.

        @param full_log whether to create full.log from the specified TestDetail
                        or a standard log
        @return the base name of the log file created
        '''

        # find log file
        logpath = self._propose_log_name(test_detail, full_log)
        assert not os.path.exists(logpath), \
            'This path should not exist: %s' % logpath

        # write the data
        with open(logpath, 'wb') as log:
            log.write(test_detail.create_output(full_log=full_log))
        print 'Log created: %s' % logpath

        return os.path.basename(logpath)


    def _propose_log_name(self, test_detail, full_log=False):
        '''Propose a file name where to write the log.

        The file is guaranteed not to exist (if @param full_log=False) and have
        some reasonable name based on TestDetail.id. For @param full_log=True
        the exact file name is always the same.

        @param full_log whether to create full.log from the specified TestDetail
                        or a standard log
        @return the full file path of the log
        '''
        if full_log:
            logname = os.path.basename(self.url['full_log'])
        else:
            # take only first 20 characters so that we don't have it too long
            logname = test_detail.id[:20] if test_detail.id else 'main'
            # sanitize the file name
            logname = logname.replace('/', '_')
            logname = logname.replace('\\', '_')
            logname = logname.replace(' ', '_')
            # ensure log doesn't exist already, generate a new name in that case
            # (by appending _1, _2, etc)
            newname = logname
            i = 1
            while os.path.exists(os.path.join(self.resultsdir, newname+'.html')):
                newname = '%s_%d' % (logname, i)
                i += 1
            logname = newname + '.html'

        logpath = os.path.join(self.resultsdir, logname)

        if not full_log:
            assert not os.path.exists(logpath), \
                'This path should not exist: %s' % logpath

        return logpath


    def log(self, message, highlight=False, output=True, printout=True,
            stderr=False, details=None, main_detail=True):
        '''Log output line (to different destinations if desired).

        By default this method prints the message to stdout and puts it to
        self.detail.output (i.e. test log file).

        In the basic case you use this method for logging everything you want
        to see in your test log file, and you don't care about @param details
        and @param main_detail. It automatically uses self.detail as the
        destination TestDetail instance where to log the message to.
        Only if you run a multi-item test and need to report more results (and
        create more logs) than just one, only then you set @param details and
        @param main_detail to log the message into a differnet TestDetail
        instance.

        There is a special log called full.log that contains everything you ever
        logged via this method (as long as @param main_detail was True). If you
        have a message that you don't want to see in your test log (e.g. it's
        too long), you have it logged into the full.log only (just set @param
        output=False).

        @param message message text to log (string)
        @param highlight highlight the message; if the value is True, then no
                         comment is provided, if the value is string, then the
                         string is provided as the comment
        @param output place the message to TestDetail.output (i.e. it will
                      appear in the log file)
                      Note: If if this is False, the message will still be
                      visible in the full.log (if @param main_detail=True).
        @param printout print the message to system console
                        This can be handy if you executed external program (its
                        output went to debug log) and now you want to reprint
                        parts of it (or in full length) into the test log. With
                        @param printout=False you won't have it duplicated in
                        the debug log.
        @param stderr print the message to stderr, otherwise to stdout; only
                      if @param printout=True
        @param details additional TestDetail instances to log this message to
                       (iterable or just a single TestDetail instance)
        @param main_detail whether to log this message to self.detail (the main
                           TestDetail object for this test)

        Examples:
        >>> log('foo')
        Log 'foo' into the test log and stdout.
        >>> log('foo', highlight=True)
        Log 'foo' into the test log and stdout and highlight it.
        >>> log('foo', stderr=True, highlight='bar')
        Log 'foo' into the test log and stderr and highlight it with comment
        'bar'.
        >>> log('foo', output=False)
        Log 'foo' into stdout but not into the test log.
        >>> log(cmd_result.stdout, printout=False)
        Reprint the command output into the test log, without printing it to the
        system console (and debug log altogether).
        >>> detail = TestDetail(test=self, id='baz-1.2-3')
        >>> log('foo', details=detail)
        Log 'foo' into the test log, stdout and also to an additionally provided
        instance of TestDetail.
        >>> log('foo', details=detail, main_detail=False)
        Log 'foo' only into the provided instance of TestDetail and stdout,
        don't log it into main test log.
        '''

        if details is None:
            details = []
        
        # store in full output
        if main_detail:
            self.detail._full_output.append(message)

        # print on console
        if printout:
            if stderr:
                print >> sys.stderr, message
            else:
                print message

        # get list of TestDetail to log to
        if not isinstance(details, collections.Iterable):
            details = [details]
        else:
            details = list(details) # copy the list because we want to modify it

        if main_detail:
            details.append(self.detail)

        # log to all TestDetails
        for detail in details:
            # send to output
            if output:
                detail.output.append(message)
            # send to highlights
            if highlight:
                comment = highlight if isinstance(highlight, str) else None
                detail.highlights.append((message, comment))


class TestDetail(object):
    '''Class encompassing everything related to the outcome of a test run.

    @attr result_order tuple of result keywords sorted by priority from least
                       important to most important
    @attr test instance of AutoQATest this TestDetail is related to
    @attr id string identification of the test. This doesn't have to be unique,
             but it should be descriptive enough for the log reader to quickly
             understand what has been tested.
    @attr summary a few words or one-sentence summary about the outcome of the
                  test run. Should not duplicate @attr id or @attr result.
    @attr highlights important lines from the test output which should be
                     highlighted. Each highlighted line can have some
                     descriptive comment assigned. It is also possible to add
                     comment without a line assigned. All these highlights will
                     be made easily visible in the resulting test log. The
                     format is [("text to highlight", "comment")]. First or
                     second item in the tuple may be empty. If the first item
                     is not empty, then the corresponding line must exist in
                     @attr output.
    @attr output output from the test run (list of strings). The test log will
                 be generated out of this. Be sure to use AutoQATest.log() to
                 populate it. You can also modify it directly (for example
                 filter out some messages you don't want to see in the test log),
                 but please note that if you add a new message directly to @attr
                 output, it won't show up in the full.log file, you have to use
                 AutoQATest.log() for that.
    @attr arch architecture for which this test was executed, or 'noarch' (default)
    @attr email_to list of email addresses to send email with results to, in
                   addition to the default email recipient defined in autoqa.conf
    @attr log instance of PrettyLog that is used to create the final log of this
              TestDetail. You can set additional attributes on this instance to
              provide further information to display in the log.
    @attr keyvals all key-value pairs in this dictionary are stored in ResultsDB
                  as 'extra data'. This is most commonly used to store the information
                  about tested envr/compose, etc.
    '''

    result_order = ('PASSED', 'INFO', 'FAILED', 'NEEDS_INSPECTION',
                    'ABORTED', 'CRASHED')


    def __init__(self, test, id, result=None, summary=None, highlights=[],
        output=[], arch='noarch', email_to=[]):
        '''Create new TestDetail.
        For param descriptions see class docstring.'''

        # validate input
        for param in (highlights, output, email_to):
            if not util.listlike(param):
                raise ValueError("'highlights', 'output' and 'email_to' "
                "parameters must be lists. Yours was: %s" % type(param))

        self.test = test
        self.id = id
        self._result = result
        self.summary = summary
        self.highlights = list(highlights)
        self.output = list(output)
        self.arch = arch
        self.email_to = list(email_to)
        self.log = PrettyLog(self)
        self.keyvals = {}
        self._resultsdb = {'testrun_id': None}

        self._full_output = list(output)
        # This is useful just for the main TestDetail, because the full.log is
        # generated out of it.


    @property
    def result(self):
        '''Keyword specifying the final result of the test. Must be one of those
        defined in self.result_order. If no result was set yet, this looks like
        NEEDS_INSPECTION.'''
        return self._result or 'NEEDS_INSPECTION'


    @result.setter
    def result(self, value):
        if value not in self.result_order:
            raise ValueError('Unknown result keyword: %s' % value)
        self._result = value


    def update_result(self, result):
        '''Update the self.result with provided value if it has higher priority
        according to self.result_order than the current value. Handles None
        values.'''
        if self.cmp_result(result, self._result) > 0:
            self.result = result


    def cmp_result(self, result1, result2):
        '''Compare two results according to self.result_order and return -1/0/1
        of the result1 has lower/equal/higher priority than result2.'''

        if (result1 not in self.result_order) and (result1 is not None):
            raise exc.AutoQAValueError('Unknown result keyword: %s' % result1)
        if (result2 not in self.result_order) and (result2 is not None):
            raise exc.AutoQAValueError('Unknown result keyword: %s' % result2)


        index1 = self.result_order.index(result1) if result1 is not None else -1
        index2 = self.result_order.index(result2) if result2 is not None else -1
        return cmp(index1, index2)


    def broken(self):
        '''Tells whether the test result is set to one of the broken states
        (i.e. ABORTED or CRASHED).'''
        return self.result in ['ABORTED', 'CRASHED']


    def create_summary_table(self):
        '''Create a pretty summary table of the test outcomes and return it
        as a string.
        '''

        summary_table = '\n'.join([
        '*'*40,
        '* TEST: %s' % self.test.__class__.__name__,
        '* RESULT: %s' % self.result,
        '* ID: %s' % self.id,
        '* SUMMARY: %s' % self.summary,
        '* ARCH: %s' % self.arch,
        '* HIGHLIGHTS: %d items' % len(self.highlights),
        '* OUTPUT: %d lines' % len('\n'.join(self.output).splitlines()),
        '* FULL OUTPUT: %d lines' % len('\n'.join(self._full_output).splitlines()),
        '* ADDITIONAL EMAIL RECIPIENTS: %s' % ', '.join(self.email_to),
        '*'*40])
        return summary_table


    def create_output(self, full_log=False):
        '''Return contents of a log (to be written, sent as email, etc).

        @param full_log whether to return contents of the full.log or the
                        standard log
        '''

        if full_log:
            # we're creating full.log
            out = '''\
%(table)s
Debug info: %(debug)s
All results: %(results)s

%(line)s
'''         % {'table': self.create_summary_table(),
               'debug': self.test.url['debug_dir'],
               'results': self.test.url['result_dir'],
               'line': 80 * '-'}
            if self.highlights:
                out += 'Highlights:\n'
                for line, comment in self.highlights:
                    out += (line.strip() if line else '--') + '\n'
                    if comment:
                        out += '\n'.join(['↳ %s' % line for line in
                                          comment.splitlines()]) + '\n'
                out += 80 * '-' + '\n'
            out += '\n'.join(self._full_output)
            return out
        else:
            # we're creating standard log
            return self.log.create()


    def excerpt(self):
        '''Make a short compilation of all the important information from
        TestDetails. The result is string like:
            "@test_name: @result; @summary for @id"
        '''

        if self.summary and self.id:
            last_part = '%s for %s' % (self.summary, self.id)
        elif self.summary:
            last_part = self.summary
        elif self.id:
            last_part = self.id
        else:
            last_part = ''

        return '%s: %s; %s' % (self.test.__class__.__name__, self.result,
                               last_part)


def create_multi_item_summary(results, result_order=TestDetail.result_order):
    '''Create a string containing a sum of all results, like this:
    "3 PASSED, 1 INFO, 2 FAILED"
    Results are sorted by @param result_order.

    @param results either one TestDetail instance or an iterable of
                        TestDetails or an iterable of result strings
    @param result_order order of the results by importance
    '''
    if isinstance(results, TestDetail):
        results = (results,)

    if len(results) <= 0:
        return ''

    if isinstance(iter(results).next(), TestDetail):
        all_results = [detail.result for detail in results]
    else: # list of strings
        all_results = results

    summary = []
    for res in result_order:
        count = all_results.count(res)
        if count > 0:
            summary.append('%d %s' % (count, res))
    summary = ', '.join(summary)

    return summary
