#
# Rawhide Acceptance Test Suite - common functions
#
# Copyright 2009, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Will Woods <wwoods@redhat.com>

import os
import sys
import gzip
import hashlib
import datetime
import urlgrabber.grabber, urlgrabber.progress
import socket
import rpmUtils, rpmUtils.miscutils, rpmUtils.arch, rpmUtils.transaction
import re
import urllib
import collections
import rpm


def get_urlgrabber(**kwargs):
    '''Get a new instance of URLGrabber that is reasonably set (progress
       bar when in terminal, automatic retries, connection timeout, etc)
       @kwargs additional kwargs to add to the URLGrabber constructor
       @return URLGrabber instance
    '''
    grabber_args = {'retry': 3, 'timeout': 120}
    grabber_args.update(kwargs)

    grabber = urlgrabber.grabber.URLGrabber(**grabber_args)

    # progress bar when in terminal
    if hasattr(sys.stdout, "fileno") and os.isatty(sys.stdout.fileno()):
        grabber.opts.progress_obj = urlgrabber.progress.TextMeter()

    # retry on socket timeout
    if 12 not in grabber.opts.retrycodes:
        grabber.opts.retrycodes.append(12)

    return grabber


grabber = get_urlgrabber()


def timestamp_to_compose_id(timestamp=None, serial=1):
    if not timestamp:
        d = datetime.datetime.today()
    else:
        d = datetime.datetime.fromtimestamp(timestamp)
    return int(d.strftime('%Y%m%d') + '%02u' % serial)

def todays_compose_id(serial=1):
    return timestamp_to_compose_id(serial=serial)

def get_basearch(arch = None):
    '''
    If 'arch' parameter is not given, this behaves like get_arch,
    but returns the basearch (as used by yum etc.)

    If 'arch' parameter is set, this function returns appropriate
    'base' arch (i.e. i386 for i[456]86, ppc for ppc64) if applicable.
    '''
    if arch == None:
        arch = os.uname()[4]
    if arch in ('i486', 'i586', 'i686'):
        arch = 'i386'
    elif arch == 'ppc64':
        arch = 'ppc'
    return arch

def makedirs(fullpath):
    '''This is the same as @method os.makedirs(), but does not throw an
    exception when the destination directory already exists.'''
    try:
        os.makedirs(fullpath)
        assert os.path.isdir(fullpath)
    except OSError, e:
        if e.errno == 17: # "[Errno 17] File exists"
            # if it is a directory everything is ok
            if os.path.isdir(fullpath):
                return
            # otherwise it is a file/socket/etc and it is an error
            else:
                raise
        else:
            raise

def ishexstring(string):
    if string.startswith('0x') or string.startswith('0X'):
        string = string[2:]
    return string.strip('abcdefABCDEF0123456789') == ''

def nonempty(filename):
    return os.stat(filename)[6] > 0

def do_gunzip(infile, outfile):
    _write_file(gzip.open(infile), open(outfile,"wb"))

def do_gzip(infile, outfile):
    _write_file(open(infile), gzip.open(outfile,'wb'))

def _write_file(infile, outfile):
    while True:
        data = infile.read(4096)
        if data:
            outfile.write(data)
        else:
            break
    infile.close()
    outfile.close()

def checkhash(filename, algo, hexdigest):
    hasher = hashlib.new(algo)
    f = open(filename)
    data = f.read(4096)
    while data:
        hasher.update(data)
        data = f.read(4096)
    hasher.update(data)
    f.close()
    r = check("%s %s" % (os.path.basename(filename), algo),
              hasher.hexdigest() == hexdigest)
    return r

def check_opt_in(pkgname, dist_release):
    """take a package name and distribution release and check for an autoqa-optin file"""

    if not pkgname or not dist_release:
        return False

    url = 'http://fedorapeople.org/packages/'
    fullurl = url + '%s/%s/%s/autoqa-optin' % (pkgname[0].lower(), pkgname, dist_release)
    tmout = socket.getdefaulttimeout()
    socket.setdefaulttimeout(10) # if it takes more than that, just consider as not opted in
    try:
        h = urllib.urlopen(fullurl)
    except IOError:
        return False
    finally:
        socket.setdefaulttimeout(tmout)

    return h.getcode() == 200 # file exists (HTTP 200 OK)


class CheckFailedException(Exception):
    pass

def check(msg, ok):
    if ok:
        print "CHECK: %s: OK" % msg
        return True
    else:
        print "CHECK: %s: FAIL" % msg
        raise CheckFailedException(msg)

def test(key, result, keyvals=None):
    if result:
        if keyvals:
            keyvals[key] = 'OK'
        print "TEST RESULT: %s: OK" % key
        return result
    else:
        if keyvals:
            keyvals[key] = 'FAIL'
        print "TEST RESULT: %s: FAIL" % key
        raise CheckFailedException(key)


def envra(rpmstr, fmt = 'envra'):
    """
    @param rpmstr string to be 'manipulated' - can be standard style rpm filename,
      envra, envr, nvr
    @param fmt desired format of the string to be returned. Allowed options are:
      envra, envr, nvra, nvr, e, n, v, r, a. If arch is not present in rpmstr,
      'noarch' is returned wherever needed.
    @returns string based on the specified format; if epoch is requested but not
      present, it is not included in the output string (rather than prepending '0:')
    """
    fmt = fmt.lower()
    supported_formats = ['envra', 'envr', 'nvra', 'nvr', 'e', 'n', 'v', 'r', 'a']
    if fmt not in supported_formats:
        raise ValueError("Format '%s' not in supported formats (%s)" % (fmt, ', '.join(supported_formats)))

    # prepare rpmstr so it can be used with rpmUtils.miscutils.splitFilename()
    parts = rpmstr.split('.')
    if parts[-1].lower() != 'rpm':
        parts.append('rpm')
    if parts[-2].lower() not in rpmUtils.arch.arches.keys() + ['noarch', 'src']:
        parts.insert(-1, 'noarch')

    # split the filename
    (n, v, r, e, a) = rpmUtils.miscutils.splitFilename('.'.join(parts))

    if len(fmt) == 1:
        return {'e':e, 'n':n, 'v':v, 'r':r, 'a':a}[fmt]

    # if epoch is not present, we do not want to return it in the
    # envr(a) requests as '0:' but rather return nvr(a) (so we take
    # advantage of `e` being empty string. This means, that if `e`
    # is non-empty, we need to add add ':' to the end, since it's
    # not present in the format string (see below)
    if e != '':
        e += ':'

    return {'envra': "%s%s-%s-%s.%s" % (e, n, v, r, a),
            'envr' : "%s%s-%s-%s" % (e, n, v ,r),
            'nvra' : "%s-%s-%s.%s" % (n, v, r ,a),
            'nvr'  : "%s-%s-%s" % (n, v, r),
           }[fmt]

def get_distro(rpmstr):
    """
    @param rpmstr string to be searched for distro tag.
      The allowed input values are the same as for envra() function.
    @returns rpmstr substring refering to the distro release
      e.g.: fc9, fc14 etc. Raises ValueError exception if no
      distro name is found
    """
    # gets just the release - filters out the package name so
    # that every 'fcXX' left will refer only to distro
    rel = envra(rpmstr,'r')
    try:
        return re.findall(r'^.*(fc\d{1,2})(\D.*)?$', rel)[-1][0]
    except IndexError:
        raise ValueError('Distro name not found in %s' % rpmstr)

def valid_rpm(rpm_file):
    '''Check that RPM file is valid.
       Sets flag for automatic md5sum checking and then tries
       to load RPM header.
       @param rpm_file RPM file path
       @return True if file seems OK or False if there is some problem (file
               can't be read, RPM header is corrupt)
    '''
    try:
        ts = rpmUtils.transaction.initReadOnlyTransaction()
        ts.setVSFlags(~(rpm.RPMVSF_NOMD5|rpm.RPMVSF_NEEDPAYLOAD))
        header = rpmUtils.miscutils.hdrFromPackage(ts, rpm_file)
    except (IOError, rpmUtils.RpmUtilsError, rpm.error), e:
        print >> sys.stderr, 'RPM verification of "%s" failed: %s' % (rpm_file, e)
        return False
    finally:
        if ts is not None:
            ts.close()

    return True


def samefile(filepath, url):
    '''Try to determine whether a local file and a file referred by HTTP URL
    are the same. Currently only their lengths are compared.
    For any exception occuring False is returned.
    '''
    try:
        # local file size
        file_size = os.path.getsize(filepath)

        # remote file size
        opener = urllib.FancyURLopener()
        remote_file = opener.open(url)
        remote_file.close()
        info = remote_file.info()
        remote_size = int(info.getheader('Content-Length'))

        return file_size == remote_size
    except:
        return False


def listlike(obj):
    '''Return whether the object is iterable (list, tuple, set, etc), but not
    string (which is an iterable, but we don't usually want to consider it as
    such).
    '''
    return isinstance(obj, collections.Iterable) and not isinstance(obj, str)


def download(url, dest, overwrite=False, grabber=grabber):
    '''Download a file.

    @param url file URL to download
    @param dest either an existing directory (string) or a file path (string).
                In the latter case all parent directories will be created if
                necessary and the last fraction of the path will be used as the
                file name.
    @param overwrite if the destination file already exists, whether to
                     overwrite or not. If False, a simple check is performed
                     whether the remote file is the same as the local file and
                     re-downloads it anyway if they are not the same.
    @param grabber custom urlgrabber.grabber.URLGrabber object
    @return file path of the downloaded file (string)
    @throw urlgrabber.grabber.URLGrabError if download fails
    '''

    # get destination file, create dirs if needed
    if os.path.isdir(dest):
        dest = os.path.join(dest, os.path.basename(url))
    else:
        makedirs(os.path.dirname(dest))

    # overwrite?
    if os.path.exists(dest):
        if overwrite:
            os.remove(dest)
        elif not samefile(dest, url):
            print ('Cached file %s differs from its online version, downloading '
                   'again' % dest)
            os.remove(dest)
        else:
            print 'Already downloaded: %s' % os.path.basename(dest)
            return dest

    print 'Downloading: %s' % url
    try:
        grabber.urlgrab(url, dest)
    except urlgrabber.grabber.URLGrabError, e:
        print >> sys.stderr, 'Download error: %s' % e
        # the file can be incomplete, remove
        if os.path.exists(dest):
            try:
                os.remove(dest)
            except:
                print >> sys.stderr, 'Could not delete incomplete file: %s' % dest
        raise

    return dest


class Tee:
    '''File-like object that redirects input into two file descriptors at once.
    @link http://shallowsky.com/blog/programming/python-tee.html'''

    def __init__(self, fd1, fd2, autoclose=True) :
        '''@param fd1 first file descriptor to write to
           @param fd2 second file descriptor to write to
           @param autoclose whether automatically try to close file descriptors
                  on this object's deletion (with the exception of sys.stdout
                  and sys.stderr)
        '''
        self.fd1 = fd1
        self.fd2 = fd2
        self.autoclose = autoclose

    def __del__(self) :
        if not self.autoclose:
            return

        for fd in [self.fd1, self.fd2]:
            if sys.stdout != fd != sys.stderr:
                if hasattr(fd, 'close'):
                    fd.close()

    def write(self, text) :
        self.fd1.write(text)
        self.fd2.write(text)

    def flush(self) :
        self.fd1.flush()
        self.fd2.flush()


def getbool(value):
    '''Converts a string into a boolean.
       Recognizes many usual strings: 1/0, yes/no, true/false, on/off; also
       empty strings.
       Throws ValueError if the string can't be recognized.
       Uses default bool conversion for non-string values.
    '''
    boolean_states = {'1': True, 'yes': True,  'true': True,   'on': True,
                      '0': False, 'no': False, 'false': False, 'off': False, '': False}

    if not isinstance(value, str):
        return bool(value)

    value = value.lower()
    if value not in boolean_states:
        raise ValueError, 'Not a boolean: %r %s' % (value, type(value))

    return boolean_states[value]

