#!/usr/bin/python
#
# Copyright 2011, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Will Woods <wwoods@redhat.com> Hongqing YANG <hoyang@redhat.com>
#
# TODO:
# - Override kickstart contents/location
#   - Overriding contents allows for crazy partitioning magic
#   - Via URL, local (virtual) media
# - Handle NFS / hdiso setup?
# - iSCSI? D:
# - Maybe convert VirtGuest to SimpleVirtGuest, and have a more complex
#   parent VirtGuest class that implements all that junk
'''simple library for creating and monitoring simple virt guests'''

import os
import sys
import time
import select

from subprocess import Popen, PIPE, call
from xml.etree import cElementTree as ElementTree

def findhostnetinfo(network, item, attr):
    '''get the attribute of the specific item of the host network'''
    p = Popen(['virsh', 'net-dumpxml', network], stdout=PIPE)
    (netxml, dummy) = p.communicate()
    assert p.returncode == 0
    root = ElementTree.fromstring(netxml)
    element = root.find(item)
    if element is None:
        return None
    for k, v in element.items():
        if k == attr:
            return v

def findhostip(network='default'):
    '''get the ip address of the specified host network'''
    return findhostnetinfo(network, 'ip', 'address')

def findhostbridge(network='default'):
    '''get the bridge name of the host network'''
    return findhostnetinfo(network, 'bridge', 'name')

class VirtGuest(object):
    '''dumb little class to hold data about virt guests'''
    def __init__(self, name, disk=None, wait=60, network='default'):
        self.name = name
        self.disk = disk or '/var/lib/libvirt/images/%s.img' % name
        self.diskcount = 1
        self.wait = wait
        self.network = network
        self.birthdate = 0
        self.showdots = False # TODO set to true if stdout is a tty?
        self.active = False
        self.create_command = None
        self.tty = None
        self.console = None
        self.pollobj = select.poll()
        self.os_variant = 'fedora15' # TODO does it influence anything?
                                     #      don't hardcode this?

    def age(self):
        '''return the age of this guest in seconds since object creation'''
        return (time.time() - self.birthdate)

    def findtty(self):
        '''get the tty console that the guest uses'''
        p = Popen(['virsh', 'ttyconsole', self.name], stdout=PIPE)
        (output, dummy) = p.communicate()
        tty = output.strip()
        assert os.path.exists(tty)
        self.tty = tty
        return tty

    def opentty(self):
        '''open tty console'''
        if self.tty is None:
            self.findtty()
        self.console = open(self.tty)
        self.pollobj.register(self.console, select.POLLIN)
        return self.tty

    def closetty(self):
        '''close tty console'''
        self.pollobj.unregister(self.console)
        self.console.close()

    def findhostip(self):
        '''get the host ip'''
        return findhostip(self.network)

    def findhostbridge(self):
        '''get the host bridge'''
        return findhostbridge(self.network)

    def poll(self, wait=None):
        '''return poll data in the specified time'''
        if wait is None:
            wait = self.wait
        return self.pollobj.poll(wait*1000)

    def poll_and_read(self, wait=None, blocksize=4096):
        '''Poll for incoming data, read it in blocksize blocks, and return it'''
        events = self.poll(wait)
        if not events:
            return None
        data = ''
        for (fd, event) in events:
            assert fd == self.console.fileno()
            if (event & (select.POLLIN | select.POLLPRI)):
                self.dot('r')
                data += os.read(fd, blocksize)
            else:
                self.dot('?')
        return data

    def dot(self, dot='.', out=sys.stdout):
        if self.showdots:
            out.write(dot)
            out.flush()

    def read_until(self, condition, readtimeout=60, max_wait=10*60, logfd=None):
        '''Read data from the guest's console until the condition becomes true
        one of the timeouts is reached.
        condition should be a callable that takes one argument - accumulated
        data read from the guest's serial console up to that point. If
        condition evaluates to True, all accumulated data will be returned.
        readtimeout is the inactivity timeout in seconds - if no data is sent
        for that amount of time, None will be returned.
        max_wait is the maximum amount of time (in seconds) to wait for the
        condition to become true - even if there's constant output.
        logfd, should be an open, writeable file descriptor. All data read from
        the serial console will be written there.
        examples:
        read_until(lambda x: 'this message indicates success' in x)
        read_until(log_parser_func, logfd=open("serial.log", "w"))
        '''
        data = ""
        start = time.time()
        while (time.time() - start) < max_wait:
            self.dot('p')
            newdata = self.poll_and_read(readtimeout)
            if newdata is None:
                break
            self.dot('w')
            if logfd is not None:
                logfd.write(newdata)
                logfd.flush()
            data += newdata
            if condition(data):
                self.dot(':D\n')
                return data
        self.dot(':/\n')

    def create(self, location, kickstart, ram=1024, disksize=10, repo=None,
               updates=None, extra_args='', serial=True):
        '''
        Create a virtual guest by calling `virt-install` with appropriate
        arguments, anaconda logs are transfered to localhost by virtio.
        @param location: A URL containing install images
        @param ram: Size of memory (in megabytes) to allocate for guest
        @param disksize: Size of disk (in gigabytes) to allocate for guest
        @param kickstart: kick start file
        @param repo: The repository location to install
        @param updates: The update images, it is a list
        @param extra_args: Additional boot arguments for the guest
        @param serial use a serial console
        @raises RuntimeError: If the return value from `virt-install` is non-zero
        '''
        if serial:
            extra_args += ' serial console=ttyS0'
        if repo:
            extra_args += ' repo=%s' % repo
        if updates:
            for update in updates:
                extra_args += ' updates=%s' % update

        virt_args = [
            '--name', self.name,
            '--ram', str(ram),
            '--vcpus', '1',
            '--os-type', 'linux',
            '--os-variant', self.os_variant,
            '--disk', 'path=%s,size=%s,sparse=false' % (self.disk, disksize),
            '--location', location,
            '--network', 'network:%s' % self.network,
            '--noautoconsole',
            '--channel', 'tcp,host=127.0.0.1:6080,mode=connect,'\
                         'target_type=virtio,name=org.fedoraproject.anaconda.log.0',
            '--graphics', 'none',
        ]

        # accept kickstart files from remote and local
        if (kickstart.startswith("http:") or kickstart.startswith("ftp:") or
            kickstart.startswith("nfs:")):
            virt_args += ['--extra-args', '%s ks=%s' % (extra_args, kickstart)]
        elif kickstart.startswith(".") or kickstart.startswith("/"):
            virt_args += ['--initrd-inject=%s' % kickstart]
            virt_args += ['--extra-args', '%s ks=file:/%s' % (extra_args,
                          os.path.basename(kickstart))]

        if self.diskcount > 1:
            for n in range(1, self.diskcount): # e.g. 5 -> [1,2,3,4]
                virt_args += ['--disk',
                  'path=%s.%d,size=%s,sparse=false' % (self.disk, n, disksize)]

        # Initiate guest install
        self.create_command = ['virt-install'] + virt_args
        retval = call(self.create_command)
        if retval != 0:
            self.delete()
            raise RuntimeError, "virt-install failed: %i" % retval
        self.birthdate = time.time()
        self.active = True

    def isocreate(self, location, ram=1024, disksize=10):
        '''
        Create a virtual guest by calling `virt-install` with appropriate
        arguments, anaconda logs are transfered to the localhost by virtio.
        @param location: the location of iso image
        @param ram: Size of memory (in megabytes) to allocate for guest
        @param disksize: Size of disk (in gigabytes) to allocate for guest
        @param extra_args: Additional boot arguments for the guest
        @raises RuntimeError: If the return value from `virt-install` is non-zero
        '''

        virt_args = [
            '--name', self.name,
            '--ram', str(ram),
            '--vcpus', '1',
            '--os-type', 'linux',
            '--os-variant', self.os_variant,
            '--disk', 'path=%s,size=%s,sparse=false' % (self.disk, disksize),
            '--cdrom', location,
            '--network', 'network:%s' % self.network,
            '--accelerate',
            '--noautoconsole',
            '--channel', 'tcp,host=127.0.0.1:6080,mode=connect,target_type=virtio,'\
                         'name=org.fedoraproject.anaconda.log.0'
        ]

        if self.diskcount > 1:
            for n in range(1, self.diskcount): # e.g. 5 -> [1,2,3,4]
                virt_args += ['--disk',
                  'path=%s.%d,size=%s,sparse=false' % (self.disk, n, disksize)]

        # Initiate guest install
        self.create_command = ['virt-install'] + virt_args
        retval = call(self.create_command)
        if retval != 0:
            self.delete()
            raise RuntimeError, "virt-install failed: %i" % retval
        self.birthdate = time.time()
        self.active = True

    def delete(self):
        '''delete the guest and the virt disk'''
        if self.name and self.disk:
            call(["virsh", "destroy", self.name])
            call(["virsh", "undefine", self.name])
            call(["virsh", "vol-delete", self.disk])
            self.active = False

    def exists(self):
        '''check the guest exist by name'''
        devnull = open(os.devnull, "w+")
        retval = call(["virsh", "dominfo", self.name],
                      stdout=devnull, stderr=devnull)
        return retval == 0

    def paircreate(self, location, ram=1024, disksize=10, kickstart=None, repo=None,
                   updates=None, extra_args='', serial=True):
        '''
        Create a virtual guest by calling `virt-install` from kernel and initrd
        pair with appropriate arguments, anaconda logs are transfered to
        localhost by virtio.
        @param location: A directory containing install images
        @param ram: Size of memory (in megabytes) to allocate for guest
        @param disksize: Size of disk (in gigabytes) to allocate for guest
        @param kickstart: kick start file, if it is None, a local kickstart
                          file will be used and upload to virtue disk
        @param repo: The repository location to install
        @param updates: The update images, it is a list
        @param extra_args: Additional boot arguments for the guest
        @raises RuntimeError: If the return value from `virt-install` is non-zero
        @serial a boolean value indicate use serial or not
        '''
        if serial:
            extra_args += ' serial console=ttyS0'
        if repo:
            extra_args += ' repo=%s' % repo
        if updates:
            for update in updates:
                extra_args += ' updates=%s' % update
        # accept kickstart files from remote and local, if kickstart file is not
        # given, use the local one
        if kickstart:
            if (kickstart.startswith("http:") or kickstart.startswith("ftp:") or
                kickstart.startswith("nfs:")):
                extra_args += ' ks=%s' % kickstart
            elif kickstart.startswith(".") or kickstart.startswith("/"):
            #FIXME upload kickstart file to virt disk
                pass
        else:
            #FIXME use locak ksfile and upload kickstart file to virt disk
            pass

        virt_args = [
            '--name', self.name,
            '--ram', str(ram),
            '--vcpus', '1',
            '--os-type', 'linux',
            '--os-variant', self.os_variant,
            '--disk', 'path=%s,size=%s,sparse=false' % (self.disk, disksize),
            '--boot', 'kernel=%s/vmlinuz,initrd=%s/initrd.img,kernel_args=%s' \
                       % (location,location,extra_args),
            '--network', 'network:%s' % self.network,
            '--accelerate',
            '--noautoconsole',
            '--channel', 'tcp,host=127.0.0.1:6080,mode=connect,target_type=virtio,'\
                         'name=org.fedoraproject.anaconda.log.0']

        if self.diskcount > 1:
            for n in range(1, self.diskcount): # e.g. 5 -> [1,2,3,4]
                virt_args += ['--disk',
                  'path=%s.%d,size=%s,sparse=false' % (self.disk, n, disksize)]

        # Initiate guest install
        self.create_command = ['virt-install'] + virt_args
        retval = call(self.create_command)
        if retval != 0:
            self.delete()
            raise RuntimeError, "virt-install failed: %i" % retval
        self.birthdate = time.time()
        self.active = True
