#!/usr/bin/python
# wiki.py - wiki interaction module for autoqa

import fedora.client
import urllib
import simplejson

from urlparse import urljoin

def get_sections(page):
    w = fedora.client.wiki.Wiki()
    params = {'action':'parse',
              'format':'json',
              'prop':'sections',
              'text':'{{:%s}}__TOC__' % page}
    result = w.send_request('api.php', req_params=params)
    return result['parse']['sections']

def get_page_section(page, secnum):
    w = fedora.client.wiki.Wiki()
    # w.send_request chokes if it doesn't get json data back, so we can't
    # use that.
    params = {'action':'raw', 'templates':'expand', 'section':secnum, 'title':page}
    url = urljoin(w.base_url, 'index.php')
    url += '?' + urllib.urlencode(params, doseq=True)
    section = urllib.urlopen(url)
    return section.read()

def wiki_json_parse(sectiondata):
    decoder = simplejson.JSONDecoder()
    jsonitems = []
    # TODO: this could be more robust. We could parse each line by itself
    # and then merge all the valid lines at the end, rather than completely
    # dying if there's any invalid data..
    for line in sectiondata.split("\n"):
        if line.startswith('*'):
            jsonitems.append(line[1:].strip())
        elif jsonitems:
            break
    jsondata = '{' + ','.join(jsonitems) + '}'
    data = decoder.decode(jsondata)
    return data

def parse_autoqa_metadata(page):
    secnum = None
    for s in get_sections(page):
        if s['line'].lower().strip() == 'autoqa metadata':
            secnum = s['number']
    if not secnum:
        return dict()
    section = get_page_section(page, secnum)
    return wiki_json_parse(section)

# --- And these do the same thing, but using pycurl

import pycurl
from fedora.client.proxyclient import _PyCurlData

def get_page_section_curl(page, secnum):
    w = fedora.client.wiki.Wiki()
    # w.send_request chokes if it doesn't get json data back, so we can't
    # use that.
    url = urljoin(w.base_url, 'index.php')
    response = _PyCurlData() # The data we get back from the server
    data = None     # decoded JSON via simplejson.load()

    request = pycurl.Curl()
    request.setopt(pycurl.URL, url)
    # Associate with the response to accumulate data
    request.setopt(pycurl.WRITEFUNCTION, response.write_data)
    # Follow redirect
    request.setopt(pycurl.FOLLOWLOCATION, True)
    request.setopt(pycurl.MAXREDIRS, 5)
    # Set standard headers
    headers = ['User-agent: %s' % w.useragent]
    request.setopt(pycurl.HTTPHEADER, headers)
    # Set the parameters
    params = {'action':'raw', 'templates':'expand', 'section':secnum, 'title':page}
    req_data = urllib.urlencode(params, doseq=True)
    request.setopt(pycurl.POSTFIELDS, req_data)
    # run the request
    request.perform()
    # FIXME if request.getinfo(pycurl.HTTP_CODE) >= 400: fail
    return response.data

def parse_autoqa_metadata_curl(page):
    secnum = None
    for s in get_sections(page):
        if s['line'].lower().strip() == 'autoqa metadata':
            secnum = s['number']
    if not secnum:
        return dict()
    section = get_page_section_curl(page, secnum)
    return wiki_json_parse(section)

if __name__ == '__main__':
    import pprint
    page = 'QA:Test_Plan_Metadata_Test_Page'
    print "Fetching metadata from %s" % page
    pprint.pprint(parse_autoqa_metadata(page))
