#!/usr/bin/python

import os
import sys
import cgi
import re
import time
import subprocess
import urlgrabber
import logging
import autoqa.logger
import autoqa.util
import autoqa.config
from autoqa.repoinfo import repoinfo
from autoqa.config import autoqa_conf
from wsgiref.validate import validator
from wsgiref.simple_server import make_server

# Set up a logger
autoqa.logger.setup_logger(logging.INFO, True)

# Setup a cache path
cachedir = os.path.join(autoqa.config.Config.paths.cachedir, 'git-post-receive')
autoqa.util.makedirs(cachedir)

def get_prevtime(cachefile):
    '''Returns the prevtime to use when looking for new builds.  Value will be
       either 1) the mtime of the cachefile, or 2) if no cachefile is found, current
       time - 3 hours.'''
    if not cachefile.startswith(cachedir):
        cachefile = os.path.join(cachedir, cachefile)
    try:
        return os.stat(cachefile)[8] # mtime
    except:
        return time.time() - 10800

def set_prevtime(cachefile):
    '''Save a list of builds for a subsequent run'''
    if not cachefile.startswith(cachedir):
        cachefile = os.path.join(cachedir, cachefile)
    out = open(cachefile,"w")
    out.close()

def index(environ, start_response):
    """This function will be mounted on "/"
    """
    # Display basic URL syntax
    data = '''
Access this app using the syntax:
   http://server/path/git-hook/REPONAME/BRANCHNAME
 for running tests for git repo REPONAME and git branch BRANCHNAME.

 Example:
   http://%s%s/git-hook/anaconda.git/master

 Optional arguments (used in the query string):
   dryrun: do not execute, just display the command(s)
''' % (environ.get('HTTP_HOST','server'), environ.get('SCRIPT_NAME','path'))
    start_response('200 OK', [('Content-Type', 'text/plain')])
    return [data]

def not_found(environ, start_response):
    """Called if no URL matches."""
    start_response('404 NOT FOUND', [('Content-Type', 'text/plain')])
    return ['Not Found']

def debug(environ, start_response):

    response = ''
    for k,v in environ.items():
        response += "<strong>%s</strong> = '%s'</br>" % (k, v)

    start_response('200 OK', [('Content-Type', 'text/html')])
    return [response]

def git_hook(environ, start_response):

    # Was a dry-run requested?
    dryrun = re.search('dry-?run', environ.get('QUERY_STRING',''))

    # Initialize response string
    response = ''

    # Parse arguments (reponame, branch)
    (name, branch) = environ.get('autoqa.url_args',('',''))

    # If no valid git repository URLs were found
    if name is None or name == '':
        response = "Ignoring request - Invalid git repo '%s'" % (name,)
        logging.warn(response)
        start_response('400 Bad Request', [('Content-Type', 'text/html')])
        return [response]

    # Default to 'master' branch
    if branch == '':
        branch = 'master'

    # Shorten the branch name (converts 'refs/heads/foo' to 'foo')
    re_group = re.match(r'(.*)/([^/]*)$', branch)
    if re_group:
        (garbage, branch) = re_group.groups()

    # Check whether repo exists
    # XXX - This check is flawed and doesn't report failure in all cases.
    #       Improve the check, or just schedule jobs
    # FIXME - This is scheduling 4 jobs, when it should be scheduling 2 (i386 and x86_64)
    git_url = ''
    if autoqa_conf.has_option('resources', 'git_baseurl'):
        for url in autoqa_conf.get('resources', 'git_baseurl').split('\n'):
            logging.debug("Checking git URL: %s" % url)
            try:
                # Determine whether this is a valid gitweb URL
                if not url.endswith('/'):
                    url = url + '/'
                urlgrabber.urlread(url + name)
            except urlgrabber.grabber.URLGrabError, e:
                # XXX - would be nice to display the URL error code
                logging.debug("Invalid git repo: %s" % url)
                git_url = ''
            else:
                logging.debug("Found valid git repo: %s" % url)
                git_url = url
                break

    # If no valid git repository URLs were found
    if git_url is None or git_url == '':
        response = "Ignoring request - Invalid git repo '%s'" % (name,)
        logging.warn(response)

    # Found a valid git repository URL
    else:
        # Check time when we last checked
        # XXX - may need to include the full URL in the cachefile to
        # distinguish between similar repos at different URL's
        cachefile = name + '|' + branch
        prevtime = get_prevtime(cachefile)
        now = time.time()
        delta = now - prevtime
        # If mtime is more than 600 seconds (10 minutes) old, do not reschedule
        # XXX -- this value should be configurable
        if delta < 600:
            response = "Ignoring request - Already triggered test %.0f " \
                       "minute(s) ago (must wait 10 minutes)." % (delta/60,)
            logging.warn(response)
        else:

            # Prepare arguments to call autoqa
            harnesscall = ['autoqa', 'git-post-receive',
                           '--name', name, '--branch', branch,
                           '--url', git_url
                          ]

            if dryrun:
                harnesscall.append('--dryrun')

            # Invoke autoqa with an --arch arg for each arch we care about
            # First, assume the arches for 'rawhide'
            testarches = repoinfo.getarches('rawhide')
            # Next, try to match the branch with a known release name
            for release in repoinfo.getreleases():
                # Will match if 'f14' in ['f14-branch', 'f14-beta-branch', ...]
                if release in branch:
                    testarches = repoinfo.getarches(release)

            # Schedule the test for each applicable arch
            if len(testarches) == 0:
                response = "Ignoring request - No applicable architectures " \
                           "found for branch %s" % branch
                logging.warn(response)
            else:
                for arch in testarches:
                    response += "%s<br>\n" % ' '.join(harnesscall + ['--arch', arch])
                    output = ''
                    try:
                        p = subprocess.Popen(harnesscall, stdout=subprocess.PIPE)
                        (output, garbage) = p.communicate()
                        rc = subprocess.call(harnesscall)
                    except OSError, e:
                        response += "%s\n" % e
                        logging.error(e)
                    else:
                        logging.info(output)

            # Cache the access time to prevent DOSing on subsequent calls
            if not dryrun:
                set_prevtime(cachefile)

    start_response('200 OK', [('Content-Type', 'text/html')])
    return [response]

def application(environ, start_response):
    """
    The main WSGI application. Dispatch the current request to
    the functions from above and store the regular expression
    captures in the WSGI environment as  `autoqa.url_args` so that
    the functions from above can access the url placeholders.

    If nothing matches call the `not_found` function.
    """

    # map urls to functions
    urls = [(r'^$', index),
            (r'^debug(.*)$', debug),
            (r'^git-hook/?([^/]+)/?(.*)$', git_hook),
           ]

    path = environ.get('PATH_INFO', '').lstrip('/')
    for regex, callback in urls:
        match = re.search(regex, path)
        if match is not None:
            environ['autoqa.url_args'] = match.groups()
            return callback(environ, start_response)
    return index(environ, start_response)

if __name__ == '__main__':
    # This is the application wrapped in a validator
    validator_app = validator(application)
    srv = make_server('', 8080, validator_app)
    try:
        print "Service running at http://localhost:8080 ..."
        srv.serve_forever()
    except KeyboardInterrupt:
        print "Exiting on user request"
