#!/usr/bin/python
#
# watch-composes.py - A utility to watch a set of compose trees for changes
#       and to kick off tests if the compose changes.
#
# Copyright 2010, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#     Jesse Keating <jkeating@redhat.com>

import os
import sys
import subprocess
import optparse
import logging
import autoqa.logger
import autoqa.config
import autoqa.util
from autoqa.repoinfo import repoinfo
from autoqa.filelock import FileLock

# Set up the option parser
parser = optparse.OptionParser(description='A utility to watch a set of \
compose trees for changes and to kick off tests if the compose changes.')
parser.add_option('--dryrun', '--dry-run', action='store_true',
    help='Do not actually execute commands, just show what would be done')
autoqa.logger.add_options(parser)
(opts, args) = parser.parse_args()

# Set up a logger
autoqa.logger.setup_logger(file_logger=not opts.dryrun, opts=opts)

# disable caching, we need always fresh data
autoqa.config.use_proxy(False)

# Lock is not created on dry-run, since dry-run does not update cachefile.
lockdir = autoqa.config.Config.paths.lockdir
lockfile = os.path.join(lockdir, 'watch-composes.lock')
lock = FileLock(lockfile)
if not opts.dryrun:
    if not lock.acquire():
        logging.warn("Another instance of this watcher is already running. Exiting...")
        sys.exit(0)

# Set the default arch to our placeholder '$ARCH'
repoinfo.setarch('$ARCH')

# Get the list of repos to watch from repoinfo (see /etc/autoqa/repoinfo.conf)
watchcomposes = [r for r in repoinfo.repos() if repoinfo.getrepo(r)['composes']]

# Setup a cache path
cachedir = os.path.join(autoqa.config.Config.paths.cachedir, 'watch-composes')
autoqa.util.makedirs(cachedir)

# Setup an empty dict of testable repos
testable = {}
repodict = {} # a place to save repo info as read

# Cycle through our list of repo watches and determine which should be tested.
for reponame in watchcomposes:
    repo = repoinfo.getrepo(reponame)
    repodict[reponame] = repo
    if not reponame in testable:
        testable[reponame] = []
    for arch in repo['arches']:
        try:
            cachefile = os.path.join(cachedir, repo['path'], arch, 'treeinfo')
            cache = open(cachefile).read()
        except IOError:
            # If we don't have a cache file, assign cache to blank.  This will
            # let us still try to hit the .treeinfo and make sure the repo is
            # alive.
            cache = ''
        # check .treeinfo
        try:
            # XXX use urljoin?
            urlpath = os.path.join(repo['url'].replace('$ARCH', arch),
                                   '.treeinfo')
            remote = autoqa.util.grabber.urlread(urlpath)
        # XXX maybe less broad exception?
        except Exception, e:
            logging.error("Error from %s: %s", urlpath, e)
            continue
        # Now compare the remote to the cache
        if cache != remote:
            testable[reponame].append(arch)

# Cycle through the testable repos and update the cache and launch harness
exit_status = 0
for reponame, arches in sorted(testable.items()):
    for arch in arches:
        repo = repodict[reponame]
        # construct arguments for call to test launcher
        harnesscall = ['autoqa', 'post-tree-compose', '--name', repo['name'],
                       '--arch', arch]
        harnesscall.append(repo['url'].replace('$ARCH', arch))
        if opts.dryrun:
            logging.info(' '.join(harnesscall))
            continue

        # update cache
        cachepath = os.path.join(cachedir, repo['path'], arch)
        autoqa.util.makedirs(cachepath)
        composepath = repo['url'].replace('$ARCH', arch)
        urlpath = os.path.join(composepath, '.treeinfo')
        output = os.path.join(cachepath, 'treeinfo')
        autoqa.util.grabber.urlgrab(urlpath, filename=output)

        # launch tests
        retval = subprocess.call(harnesscall)
        if retval != 0:
            exit_status = 10

lock.release()

sys.exit(exit_status)
