#!/usr/bin/python -tt
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# copyright (c) 2008 Red Hat, Inc
#
# Authors: Seth Vidal and Jesse Keating

"""Go through a yum packagesack and return the list of newest pkgs which
   possibly have conflicting files"""


import yum
import yum.Errors
import yum.misc
import datetime
import time
import sys
import os
import os.path
import tempfile
import shutil
import urllib
from collections import defaultdict
from rpmUtils.miscutils import rangeCompare as rpmRangeCompare

from optparse import OptionParser

# iterate all pkgs, build a filedict of filename = [list of pkg objects]
# then go through that list, if any file has more than one pkg then put those
# pkgs in another list
# print out the file conflict and the pkgs which potentially conflict
# later: work on whittling down the list by removing:
#        pkgs which are multilib sets
#        pkgs which have fully matching overlapping files (md5sums + timestamps)

def parseArgs():
    usage = """
    Check given repos for possible file/package conflicts

    %s [-c <config file>] [-r <repoid>] [-r <repoid2>]
    """ % sys.argv[0]
    parser = OptionParser(usage=usage)
    parser.add_option("-c", "--config", default='/etc/yum.conf',
        help='config file to use (defaults to /etc/yum.conf)')
    parser.add_option("-r", "--repoid", default=[], action='append',
        help=("specify repo ids to query, can be specified multiple times "
             "(default is all enabled)"))
    parser.add_option("-t", "--tempcache", default=False, action="store_true",
        help="Use a temp dir for storing/accessing yum-cache")
    parser.add_option("--repofrompath", action="append",
                      help=("specify repoid & paths of additional repositories "
        "- unique repoid and complete path required, can be specified multiple "
        "times. Example. --repofrompath=myrepo,/path/to/repo"))
    (opts, args) = parser.parse_args()
    return (opts, args)


# not a great check but it does toss out all the glibc.i686 vs glibc.i386 ones
def pkg_names_match(pkglist):
    p = pkglist[0]
    for pkg in pkglist[1:]:
        if p.name != pkg.name:
            return False

    return True


# if they have the same sourcerpm it is highly doubtful it is a legit conflict
def sourcerpms_match(pkglist):
    p = pkglist[0]
    for pkg in pkglist[1:]:
        if p.sourcerpm != pkg.sourcerpm:
            return False

    return True


# XXX we should just be able to use the pkg.inPrcoRange method but I think it's
# broken - it seems to have the rangeCompare arguments backwards.
def inPrcoRange(prcotype, p1, p2):
    p2_tuple = (p2.name, 'EQ', (p2.epoch, p2.ver, p2.rel))
    for obs in p1.returnPrco(prcotype):
        if rpmRangeCompare(obs, p2_tuple):
            return True
    return False


def check_list_for_prco(prcotype, pkglist):
    '''Compare each pair of packages in the pkglist to see if any of them
    obsoletes/conflicts/requires/provides the other'''
    for p1 in pkglist:
        for p2 in pkglist:
            if p1 == p2:
                continue
            if inPrcoRange(prcotype, p1, p2) or inPrcoRange(prcotype, p2, p1):
                #print '%s %s %s' % (p1, prcotype, p2)
                return True
    return False


def obsolete_each_other(pkglist):
    return check_list_for_prco('obsoletes', pkglist)


def package_conflict(pkglist):
    return check_list_for_prco('conflicts', pkglist)


# XXX with sufficiently fast network, the bottleneck is CPU in the checksum
# comparison here. Maybe we should multithread it someday?
def file_conflict(fn, pkglist):
    checksum = None

    for pkg in pkglist:
        #print "Fetching Headers for %s" % pkg
        #download header
        # make this into a proper error func check for urlgrabber routines
        # so we get the proper retries
        if not os.path.exists(pkg.localHdr()):
            try:
                hdr = pkg.returnLocalHeader()
            except yum.Errors.RepoError, e:
                pkg.repo.getHeader(pkg)
        else:
            pkg.repo.getHeader(pkg)
        #read in header
        try:
            hdr = pkg.returnLocalHeader()
        except yum.Errors.RepoError, e:
            #print "Cannot retrieve header for %s" % pkg
            return True

        fi = hdr.fiFromHeader()
        for fitup in fi:
            if fitup[0] == fn:
                if checksum is None:
                    checksum = fitup[-1]
                elif checksum != fitup[-1]:
                    return True

    return False


(opts, cruft) = parseArgs()

my = yum.YumBase()
my.doConfigSetup(fn = opts.config, init_plugins=False)

temp_cachedir = None
if os.geteuid() != 0 or opts.tempcache:
    temp_cachedir = tempfile.mkdtemp(prefix='conflicts')
    my.conf.cachedir = temp_cachedir
else:
    my.conf.cachedir = yum.misc.getCacheDir()

my.repos.setCacheDir(my.conf.cachedir)

if opts.repofrompath:
# setup the fake repos
    for repo in opts.repofrompath:
        repoid, repopath = tuple(repo.split(','))
        if repopath[0] == '/':
            baseurl = 'file://' + repopath
        else:
            # This URL may be a redirect, like download.fedoraproject.org. We
            # need to resolve this redirect now, otherwise the redirect URL will
            # be used for every package header download and that a) will extremely
            # slow down download speed b) may cause consistency problems if not
            # always the same repo mirror is returned from the redirect
            net_obj = urllib.urlopen(repopath)
            baseurl = repopath = net_obj.geturl()

        repopath = os.path.normpath(repopath)
        newrepo = yum.yumRepo.YumRepository(repoid)
        newrepo.name = repoid
        newrepo.baseurl = baseurl
        newrepo.basecachedir = my.conf.cachedir
        newrepo.metadata_expire = 0
        newrepo.timestamp_check = False
        my.repos.add(newrepo)
        my.repos.enableRepo(newrepo.id)
        my.logger.info("Added %s repo from %s" % (repoid, repopath))

if opts.repoid:
    for repo in my.repos.repos.values():
        if repo.id not in opts.repoid:
            repo.disable()
        else:
            repo.enable()

fulldict = defaultdict(list)
#dirdict = defaultdict(list)

print 'Getting complete filelist for:'
for repo in my.repos.listEnabled():
    print repo.urls[0]

for pkg in my.pkgSack.returnNewestByNameArch():
    for fn in pkg.filelist:
        fulldict[fn].append(pkg)

#    for fn in pkg.dirlist:
#        dirdict[fn].append(pkg)
print '%u files found.\n' % len(fulldict)

print 'Looking for duplicated filenames:'
filedict = {}
for fn in fulldict.keys():
    if len(fulldict[fn]) >= 2:
        filedict[fn] = fulldict[fn]
del fulldict # frees up a little memory
print '%u duplicates found.\n' % len(filedict)

print 'Doing more advanced checks to see if these are real conflicts:'
count = 0
found = 0
total = len(filedict)
milestone = int(0.05 * total)
start = time.time()
last = start
package_conflicts = set()
confdict = {}
for (fn, pkglist) in filedict.iteritems():
    #print '%d checking on %s ...' % (count, fn),
    if (pkg_names_match(pkglist) or
        sourcerpms_match(pkglist) or
        obsolete_each_other(pkglist)):
        pass
    elif package_conflict(pkglist):
        strlist = sorted([str(p) for p in pkglist])
        package_conflicts.add('\n'.join(strlist))
    elif file_conflict(fn, pkglist):
        found += 1
        confdict[fn] = pkglist

    # Timing / counting output
    count += 1
    if count % milestone == 0:
        now = time.time()
        percent = round(float(count*100) / total)
        files_per_sec = float(milestone) / (now - last)
        total_files_per_sec = float(count) / (now - start)
        eta_sec = float(total - count) / total_files_per_sec
        eta = str(datetime.timedelta(seconds=int(eta_sec)))
        print "%3u%% complete (%6u/%6u, %5u/sec), %4u found - eta %s" % \
            (percent, count, total, files_per_sec, found, eta)
        last = now
del filedict
print "%u file conflicts found." % len(confdict)
print "%u package conflicts found." % len(package_conflicts)

# Reduce the results to a dict like {"pkga,pkgb": [conflicting files], ...}
rbpp = defaultdict(list)# Results By Package-Pair
for (fn, pkglist) in confdict.iteritems():
    pkgpair = '\n'.join(sorted([str(p) for p in pkglist]))
    rbpp[pkgpair].append(fn)
del confdict

print '\n== Package conflicts =='
for pkglist in package_conflicts:
    print pkglist + "\n"

print '\n== File conflicts, listed by conflicting packages =='
for (pkgpair, files) in rbpp.iteritems():
    print pkgpair
    for f in sorted(files):
        print '  ' + f
    print

# delete cache
if temp_cachedir:
    shutil.rmtree(temp_cachedir)
