#
# Copyright 2010, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Will Woods <wwoods@redhat.com>

# Notice: Most recent documentation is available at doc/test_class.py.template.

import os
import re
import pickle
import tempfile
import urlgrabber
import autoqa.util
import autoqa.config
import autoqa.test
from autoqa.test import AutoQATest, TestDetail
from autoqa.decorators import ExceptionCatcher
from autotest_lib.client.bin import utils
from autoqa.repoinfo import repoinfo
from autoqa.bodhi_utils import bodhi_already_commented, query_update, build2update
from autoqa.koji_utils import SimpleKojiClientSession
from depcheck_pretty import DepcheckPretty

# --- Everything below here is depcheck-specific

class depcheck(AutoQATest):

    @ExceptionCatcher()
    def setup(self, *args, **kwargs):
        retval = utils.system('yum -y install python-rpmfluff mash')
        assert retval == 0


    @ExceptionCatcher()
    def initialize(self, config, arch, kojitag, **kwargs):
        self.is_multiresult = True
        super(self.__class__, self).initialize(config)
        self.detail.id = kojitag + '-pending'
        self.detail.arch = arch
        # dict of problematic NVRs - can't find in Bodhi
        # nvr -> error
        self.problematic = dict()
        self.koji = SimpleKojiClientSession()
        self.koji.check_connection()


    def is_accepted(self, nvr):
        '''Return True if the given nvr has been marked as Accepted by a
        previous depcheck run, False otherwise.'''
        try:
            (result, time) = bodhi_already_commented(nvr, 'depcheck',
                                                     self.detail.arch)
            return result == 'PASSED'
        except ValueError as e:
            msg = 'ERROR: %s' % e
            self.log(msg, highlight = True)
            self.problematic[nvr] = e
            return False


    @ExceptionCatcher()
    def run_once(self, nvrs, kojitag, arch, **kwargs):
        super(self.__class__, self).run_once()

        # Get our inputs
        pending_tag = kojitag + '-pending'
        self.log('Koji tag to be tested: %s' % pending_tag)
        self.log('Querying Koji for builds...')
        builds =  self.koji.listTagged(pending_tag)
        pending = sorted(set([build['nvr'] for build in builds]))
        self.log('Builds to be tested:\n  %s' % '\n  '.join(pending))

        self.detail.keyvals['kojitag'] = pending_tag

        # variable build2update contains only 'ok' builds
        self.log('Querying Bodhi for matching updates...')
        build_to_update, failures = build2update(pending, strict = True)

        # pending_ok contains just builds which passed the 'strict'
        # requirements of build2update. I.e. this is all pending builds
        # minus builds not found in Bodhi and builds from 'incomplete updates'
        # where 'incomplete update' means that not all builds from that
        # update were scheduled in Bodhi update.
        pending_ok = set(build_to_update.keys())

        for build, update in failures.items():
            if update is None:
                # build not found in Bodhi
                msg = 'Update containing %s not found in Bodhi' % build
            else:
                # some other builds from the matching update were not scheduled
                msg = ("Build %s is part of an incompletely scheduled Bodhi "
                       "update '%s'." % (build, update['title']))
            self.log(msg, highlight=True, stderr=True)
            self.problematic[build] = msg

        self.log('Bodhi updates found:\n  %s' % '\n  '.join(
            sorted(set([update['title'] for update in build_to_update.values()]))))

        self.log('Querying Bodhi for accepted builds...')
        # XXX set testarch to noarch for noarch updates?
        accepted_builds_nvrs = set([nvr for nvr in pending_ok
                                    if self.is_accepted(nvr)])
        self.log('Accepted builds:\n  %s' % '\n  '.join(sorted(accepted_builds_nvrs)))
        pending_builds_nvrs = sorted(pending_ok.difference(accepted_builds_nvrs).\
                                     difference(self.problematic.keys()))
        self.log('Pending builds:\n  %s' % '\n  '.join(pending_builds_nvrs))
        self.log('Problematic builds:\n  %s' % '\n  '.join(self.problematic.keys()))

        # if there are no pending builds, there is no sense in downloading
        # packages and running depcheck. Just summarize the decision and
        # exit gracefully
        if len(pending_builds_nvrs) == 0:
            self.log('No pending builds found. Nothing to do.',
                     highlight = True, stderr = True)
            self.detail.summary = "No pending builds. Nothing to do."
            # if we have problematic builds, print them and the errors out
            # and set the results to 'NEEDS_INSPECTION'
            if len(self.problematic) > 0:
                self.print_problematic()
                self.detail.update_result('NEEDS_INSPECTION')
            else:
                self.detail.update_result('PASSED')

            return

        # Fetch packages and build commandline
        repo = repoinfo.getrepo_by_tag(kojitag)
        repoid = repo['name']
        cmd = './depcheck %s' % repoid
        cmd_accepted = []
        for (nvr, rpms) in self.fetch_nvrs(accepted_builds_nvrs):
            for rpm in rpms:
                cmd_accepted.append(rpm)

        # rpm_to_build keys() is a set of all downloaded rpms (stripped from
        # /path/OpenThreads-2.8.3-10.fc15.i686.rpm to OpenThreads-2.8.3-10.fc15).
        # Value for the key is then nvr of the build containing this rpm file.
        #
        # NOTE: Please note, that this is something completely different than
        #       build_to_update variable which maps build to update, not rpm to build.
        #
        # Example:
        # --------
        # Fetching RPMs for OpenSceneGraph-2.8.3-10.fc15
        #    OpenSceneGraph-2.8.3-10.fc15.i686.rpm
        #    OpenThreads-2.8.3-10.fc15.i686.rpm
        #
        # rpm_to_build = {OpenSceneGraph-2.8.3-10.fc15 : OpenSceneGraph-2.8.3-10.fc15,
        #                 OpenThreads-2.8.3-10.fc15 : OpenSceneGraph-2.8.3-10.fc15}
        rpm_to_build = {}
        build_to_rpms = {}
        cmd_test = []
        for (nvr, rpms) in self.fetch_nvrs(pending_builds_nvrs):
            build_to_rpms[nvr] = map(os.path.basename, list(rpms))
            for rpm in rpms:
                cmd_test.append(rpm)
                # While we're here, construct subpackage NVR -> parent NVR map
                # package filenames are all: /some/path/to/N-V-R.A.rpm
                this_nvr = autoqa.util.envra(os.path.basename(rpm), 'nvr')
                rpm_to_build[this_nvr] = nvr

        # Store the previously accepted / to_test to files
        fd = open(os.path.join(self.resultsdir, "previously_accepted_builds.txt"), 'w')
        fd.write('\n'.join(cmd_accepted))
        fd.close()

        fd = open(os.path.join(self.resultsdir, "pending_builds.txt"), 'w')
        fd.write('\n'.join(cmd_test))
        fd.close()

        # Run the test
        fd = tempfile.NamedTemporaryFile(delete=False)
        pickle.dump({'accepted': cmd_accepted, 'to_test': cmd_test}, fd)
        fd.close()

        cmd += " -f %s" % fd.name #path to the temp file
        depcheck_output = utils.system_output(cmd, retain_output=True)

        self.log(depcheck_output, printout = False)
        digest = DepcheckPretty(depcheck_output, build_to_rpms)

        os.unlink(fd.name)

        # variable 'results' is dict containing three keys IGNORE, REJECT, ACCEPT
        # value for each of the keys is list of rpm files
        # ignored/rejected/accepted by the depcheck test.
        #
        # Note that the '.rpm' is removed from the file name.
        # The dict then could be e.g.
        #
        # results = {'ACCEPT': [OpenSceneGraph-2.8.3-10.fc15.i686],
        #            'REJECT': [OpenThreads-2.8.3-10.fc15.i686]}

        # Process output
        results = dict(IGNORE=[], REJECT=[], ACCEPT=[])
        data_re = re.compile('^(IGNORE|REJECT|ACCEPT): (.*)$')
        for line in depcheck_output.splitlines():
            match = data_re.match(line)
            if not match:
                continue
            (what, builds) = match.groups()
            results[what] = builds.split()

        # sets accepted_builds, rejected_builds, ignored_builds map the
        # rpm filenames to the build names which contain the respective
        # rpm file (taken from the rpm_to_build).
        # One example better than thousand words:
        #
        # results = {'ACCEPT': [OpenSceneGraph-2.8.3-10.fc15.i686],
        #            'REJECT': [OpenThreads-2.8.3-10.fc15.i686]}
        #
        # accepted_builds = set([OpenSceneGraph-2.8.3-10.fc15])
        # rejected_builds = set([OpenSceneGraph-2.8.3-10.fc15])
        #
        # (because both rpms are part of the OpenSceneGraph-2.8.3-10.fc15
        # build).

        accepted_builds = set()
        rejected_builds = set()
        ignored_builds = set()
        for (set_, keyword) in [(accepted_builds, 'ACCEPT'),
                                 (rejected_builds, 'REJECT'),
                                 (ignored_builds, 'IGNORE')]:
            for envra in results[keyword]:
                nvr = autoqa.util.envra(envra, 'nvr')
                if rpm_to_build.has_key(nvr):
                    set_.add(rpm_to_build[nvr])
                else:
                    self.log("Build in %s but not in pending updates, skipping: "
                             "%s" % (keyword, envra), stderr = True)

        # Now we need inspect accepted_builds, and discard the builds which
        # are also present in the rejected_builds (or ignored_builds) set.
        #
        # This is because depcheck outputs accepted/rejected rpms, which
        # we then map to builds. And if you look at the previous examples
        # it's apparent that since one rpm from the build passed, and
        # the second failed, the build's nvr is in both sets.
        #
        # But the build should pass only if all the rpms passed.
        #
        # After this step, accepted_builds and rejected_builds should be:
        #
        # accepted_builds = set([])
        # rejected_builds = set([OpenSceneGraph-2.8.3-10.fc15])

        accepted_builds = accepted_builds.difference(rejected_builds)
        accepted_builds = accepted_builds.difference(ignored_builds)

        # Now let's sanity check depcheck output. Every pending build passed
        # to depcheck on command line must be either in accepted_builds, or
        # rejected builds, or ignored builds. If we passed a build to depcheck
        # and it didn't printed it out in the results, something is very wrong
        # and we want to cancel the whole test.

        missing_builds = [build for build in pending_builds_nvrs if
                                build not in accepted_builds and
                                build not in rejected_builds and
                                build not in ignored_builds]

        # This code simply takes all builds, and creates a dictionary
        # of update metadata (taken from bodhi).
        # Because as one build can contain multiple rpms, one update
        # can contain multiple builds. This groups builds into updates.
        # The variable 'updates' then looks like
        #
        # updates = {updatename : bodhi_metadata}

        # FIXME: this should be removed once #284 and #325 are fixed
        # we want to not report results on updates with missing builds
        ignored_updates = set()

        updates = {}
        all_builds = set(rpm_to_build.values())
        for nvr in all_builds:
            bodhi_info = query_update(nvr)
            if bodhi_info is None:
                self.log("Update containing %s not found in Bodhi" % nvr,
                         stderr = True, highlight = True)
                continue

            ######################################################################
            # FIXME:
            # This is a quick, sub-optimal fix to mitigate the effects of missing
            # build information as described in #284 and #325
            # Eventually, this needs to be fixed properly.
            #######################################################################

            # if build was previously marked as ignored, move on to the next one
            if bodhi_info['title'] in ignored_updates:
                continue

            # if build is missing, mark update as ignored
            if nvr in missing_builds:
                ignored_updates.add(bodhi_info['title'])

                # make sure that the update isn't going to be posted
                if bodhi_info['title'] in updates.keys():
                    del updates[bodhi_info['title']]

                continue

            if bodhi_info['title'] not in updates.keys():
                updates[bodhi_info['title']] = bodhi_info

        # Because bodhi is working on top of updates, we need to check whether
        # all builds from the respective update were accepted. If at least
        # one of the builds in update was in rejected/ignored set, whole
        # update is 'failed'.
        #
        # If at least one update in the whole -pending repo is failed,
        # the overall depcheck result is FAILED.

        update_result = dict() # {update_title: TestDetail}
        passed_updates = 0
        failed_updates = 0
        for update in updates.values():
            td = TestDetail(self, id=update['title'], arch=arch)
            td.result = 'PASSED'
            td.log.kojitag = pending_tag
            passed_builds = 0
            failed_builds = 0
            update_result[update['title']] = td

            desc = 'Update %(result)s: %(title)s\n'

            for build in update['builds']:
                # ticket 412
                # Sometimes, update is changed during depcheck testrun
                # and new build (which was not in -pending at the start)
                # is added. When this happens, we want to skip the build
                # and mark it as problematic (to subsequently fail the 
                # update)
                try:
                    td.output.append(digest.digest[build['nvr']]['all'])
                except KeyError:
                    msg = "Build %s is part of update '%s', but it was " \
                          "added during the depcheck run" % (build['nvr'], update['title'])
                    self.log(msg, stderr = True, highlight = True, details = td)
                    td.update_result("ABORTED")
                    self.problematic[build['nvr']] = msg
                    continue

                td.output.append("\n\n")
                for h in digest.digest[build['nvr']]['highlights']:
                    td.highlights.append((h, None))

                if build['nvr'] not in pending_ok:
                    msg = "Build %s is part of update '%s', but it wasn't " \
                          "scheduled for testing" % (build['nvr'], update['title'])
                    self.log(msg, stderr = True, highlight = True, details = td)
                    td.update_result('ABORTED')
                    self.problematic[build['nvr']] = msg
                    continue

                if (build['nvr'] not in accepted_builds) and (build['nvr'] not in ignored_builds):
                    td.update_result('FAILED')
                    self.detail.update_result('FAILED')
                    failed_builds += 1
                else:
                    passed_builds += 1

                desc += self.build_status(build['nvr'], rpm_to_build,
                        accepted_builds, rejected_builds, ignored_builds)

            desc = desc % {'title': update['title'], 'result': td.result}
            desc = desc.rstrip() # strip last linefeed
            self.log(desc)
            td.log.summary_detail = desc

            td.summary = autoqa.test.create_multi_item_summary(
                ['PASSED']*passed_builds + ['FAILED']*failed_builds)

            if td.result == 'PASSED':
                passed_updates += 1
            else:
                failed_updates += 1

            # We need to set-up the keyval-pairs for the resultsdb purposes
            td.keyvals.update({'update': update['title'], 'envr': [build['nvr'] for build in update['builds']], 'kojitag': pending_tag})



        # FIXME: this should probably go away when #284 and #385 get fixed
        # display information about missing build information
        if missing_builds:
            self.log('Some builds passed in as pending, but depcheck produced '
                     'no result for them: %s' % ' '.join(missing_builds),
                     stderr=True, highlight=True)
            self.detail.update_result('NEEDS_INSPECTION')

        # Report all results to Bodhi
        for title, td in update_result.items():
            self.post_results(td, bodhi = {'title': title})

        # FIXME: This needs to be removed once #284 and #325 are fixed
        if len(ignored_updates) > 0:
            self.log('Some updates were ignored because of some problem, '
                     'setting status to NEEDS_INSPECTION', stderr=True,
                     highlight=True)
            self.detail.update_result('NEEDS_INSPECTION')

        if len(self.problematic) > 0:
            self.print_problematic()
            self.detail.update_result('NEEDS_INSPECTION')

        # TODO: force filling up reports for updates (resultsdb stuff)

        summary = autoqa.test.create_multi_item_summary(update_result.values())
        if not summary:
            summary = "0 PASSED, 0 FAILED"
        self.detail.summary = summary
        self.detail.update_result('PASSED')


    def print_problematic(self):
        '''Print out all problematic builds and their reasons'''
        msg = "PROBLEMATIC BUILDS FOUND!\n"
        for build in self.problematic.keys():
            msg += '  %s - %s\n' % (build, self.problematic[build])
        self.log(msg, highlight = True, stderr = True)


    def build_status(self, build_nvr, rpm_to_build, accepted_builds, rejected_builds, ignored_builds):
        build_rpms = set([key for key, value in rpm_to_build.items() if value == build_nvr])
        desc = '  Build %(result)s: %(build)s\n'

        rej = ""
        for rpm in build_rpms.intersection(rejected_builds):
            rej += '      %s\n' % rpm
        ign = ""
        for rpm in build_rpms.intersection(ignored_builds):
            ign += '      %s\n' % rpm

        if rej:
            desc += '    Rejected:\n'
            desc += rej
        if ign:
            desc += '    Ignored:\n'
            desc += ign

        result = 'FAILED' if ((build_nvr not in accepted_builds) and (build_nvr not in ignored_builds)) else 'PASSED'
        desc = desc % {'result': result, 'build': build_nvr}
        return desc

    def fetch_nvrs(self, builds):
        '''Fetch all the RPMs for the given builds.
        @param builds list of NVRs of builds
        @return list of tuples (build nvr,list of rpm file paths)'''
        # NOTE: Currently we're just fetching everything from koji. This might not
        # be the fastest thing, but it works for everyone, everywhere. If there's
        # a shortcut we can use on the autoqa test systems, we should use that,
        # but it needs to gracefully fall back to something like this.

        # download only necessary architectures
        arch = autoqa.util.get_basearch()
        assert arch in ['i386', 'x86_64']
        if arch == 'x86_64':
            arches = ['x86_64', 'i386', 'noarch'] # needed for multilib
        else: # i386
            arches = ['i386', 'noarch']

        rpms = []
        for nvr in sorted(builds):
            # XXX deal with debuginfo/src RPMs?
            try:
                pkgs = self.koji.get_nvr_rpms(nvr, self.tmpdir,
                       arches=arches, debuginfo=False, src=False)
            except urlgrabber.grabber.URLGrabError as e:
                self.detail.update_result('ABORTED')
                raise
            print
            rpms.append((nvr, pkgs))
        return rpms

