# depcheck_lib.py: all the actual heavy-lifting depcheck code
#
# Copyright 2010, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Will Woods <wwoods@redhat.com>

# The overall strategy - make Yum assume that all the packages from the given
# repo(s) are installed, and try to resolve deps for the new packages as if
# they were being installed.
#
# As geppetto suggests, we use two YumBase objects. First, we set one up with
# all the requested repos, then use its pkgSack (which contains package
# objects for everything in the repos) to populate the rpmdb of the second
# YumBase object. This makes it look like everything available in the repos
# is installed on the system. We then attempt to apply the new packages as an
# update to that system.
#
# But! NOTE! Using this as a simple pass/fail test will not suffice. We need to
# return a (possibly empty) set of packages which *do* pass depchecking -
# something like what yum --skip-broken does. Here's an example that shows why:
# Assume we have some packages {A,B,C,D,...} where A requires B, C requires D,
# and all other items are independent. If we use depcheck as a pass/fail test
# and we test the following sets of packages - like we might test things as
# they came out of the build system:
# 1. {A} -> FAIL
# 2. {A,C} -> FAIL
# 3. {A,C,B} -> FAIL
# 4. {A,C,B,E,F,G,...} -> FAIL
# 5. {A,C,B,E,F,G,...,D} -> PASS
# In step 4, even though the subset {A,B,E,F,G,...} would pass depcheck,
# we reject the entire set because of the presence of C. No good!
# Therefore we need to return lists of passing/failing packages instead:
# 1. {A} -> {}
# 2. {A,C} -> {}
# 3. {A,C,B} -> {A,B}
# 4. {A,C,B,E,F,G,...} -> {A,B,E,F,G,...}
# 5. {A,C,B,E,F,G,...,D} -> {A,B,C,D,...}
# So we're using _skipPackagesWithProblems to throw out the failing packages
# and make a list of passing packages.

import os
import sys
import yum
import glob
import tempfile
import subprocess
import rpmUtils.arch as rpmarch
from autoqa.repoinfo import repoinfo
from yum.parser import varReplace
from rpmUtils.miscutils import compareEVR

sys.path.append('/usr/share/yum-cli')
from cli import YumBaseCli

# TODO Should use stuff in (e.g.) rpmUtils.arch to generate this list.
# TODO Support non-Intel arches, too.
both_arches = ['i386', 'x86_64']

def prefilter_rpmdb(pkgSack):
    """
    Takes pkgSack and filters it in a way, that it contains
    only the latest version of each package.
    NOTE that this alters the pkgSack passed in arg!
    """
    # best = dictionary
    # keys = '%s_%s' % (pkg.name, pkg.arch)
    # values = pkgSack.package object
    best = {}
    for package in pkgSack.simplePkgList():
        NA = '%s_%s' % (package[:2]) # name_arch
        try: # check wheter 'best' already contains the name/arch combination
            old = best[NA]
        except KeyError: #no previous intro
            best[NA] = package
            old = best[NA]
        else: #previous info for that package & arch combo, compare
            if compareEVR((old[2] or '0', old[3], old[4]), (package[2] or '0', package[3], package[4])) < 0: # old_EVR < package_EVR
                best[NA] = package
    # now we have the list of packages we want to keep, so
    # let us create the supplement of the whole pkglist (= packages to remove)
    old_pkg_tuples = pkgSack.simplePkgList()
    for package in best.values():
        old_pkg_tuples.remove(package)
    # remove all the 'old' versions from pkgSack
    for package in old_pkg_tuples:
        pkgSack.delPackage(pkgSack.searchPkgTuple(package)[0])


class YumDepcheck(YumBaseCli):
    def __init__(self):
        YumBaseCli.__init__(self)
        # disable all plugins
        self.preconf.init_plugins = False
        # hack to prevent YumDepcheck from setting up system repos
        self.conf.config_file_path = self.conf.reposdir = '/var/empty'

    # Copy and pasted (and modified) from YumBase - see
    # /usr/lib/python*/site-packages/yum/__init__.py
    # Modified to not traceback if there's no 'yumdb_info' attribute, since
    # we're lying about the packages being installed (and thus having some
    # data in the yumdb)
    def _add_up_txmbr(self, requiringPo, upkg, ipkg):
        txmbr = self.tsInfo.addUpdate(upkg, ipkg)
        if requiringPo:
            txmbr.setAsDep(requiringPo)
        if hasattr(ipkg, 'yumdb_info') and \
           ('reason' in ipkg.yumdb_info and ipkg.yumdb_info.reason == 'dep'):
            txmbr.reason = 'dep'
        return txmbr

def do_mash(pkgdir, arches, distname=None):
    '''Set up a mash object with an ad-hoc config'''
    import mash, mash.config, StringIO
    from ConfigParser import RawConfigParser
    if distname is None:
        distname = os.path.basename(pkgdir)
    conf = mash.config.readMainConfig('/etc/mash/mash.conf')
    dist = mash.config.MashDistroConfig()
    parser = RawConfigParser()
    mash_conf = '''[%s]
output_subdir = %s
rpm_path = %s
strict_keys = False
multilib = True
multilib_method = devel
tag = this-is-a-dummy-tag
inherit = False
debuginfo = False
delta = False
source = False
arches = %s
keys = 
''' % (distname, distname, pkgdir, arches)
    parser.readfp(StringIO.StringIO(mash_conf))
    dist.populate(parser, distname, conf)
    conf.distros.append(dist)
    dist.repodata_path = dist.rpm_path
    dist.name = distname
    themash = mash.Mash(dist)
    # HACK - keep mash from logging excessively
    themash.logger.handlers = [themash.logger.handlers[-1]]
    # gotta do createrepo to get mash to Do The Right Thing
    subprocess.call(['createrepo', pkgdir])
    rc = themash.doMultilib()
    conf.distros.pop()
    return rc

def set_up_yum_object(repoid):
    y = YumDepcheck()
    # yum.misc.getCacheDir() gives us a temporary cache dir
    # TODO: use a non-temp dir so we can use cached data!
    y.conf.cachedir = yum.misc.getCacheDir()
    y.repos.setCacheDir(y.conf.cachedir)

    # Set up repo objects for requested repo and its parents
    for r in repoinfo.getparents(repoid) + [repoid]:
        repo = repoinfo.getrepo(r)
        newrepo = yum.yumRepo.YumRepository(r)
        newrepo.name = r
        baseurl = varReplace(repo['url'], y.conf.yumvar)
        newrepo.baseurl = baseurl
        newrepo.basecachedir = y.conf.cachedir
        newrepo.metadata_expire = 0
        newrepo.timestamp_check = False
        y.repos.add(newrepo)
        y.repos.enableRepo(newrepo.id)
        y.logger.info("Added repo: %s" % r)
    # We're good - return the Yum object
    return y

class MetaSackRPMDB(yum.packageSack.MetaSack):
    '''Extend MetaSack to simulate an RPMDB that has all the listed packages
    installed.'''
    def __init__(self, metasack):
        yum.packageSack.MetaSack.__init__(self)
        for (k, v) in metasack.__dict__.iteritems():
            self.__dict__[k] = v
        self._cached_conflicts_data = None
        self.__cache_rpmdb__ = False

    def fileRequiresData(self):
        '''A real rpmdb has this method, but a MetaSack doesn't. Let's fake it.
        (See /usr/lib/python*/site-packages/yum/rpmsack.py:fileRequiresData)
        '''
        installedFileRequires = {}
        installedUnresolvedFileRequires = set()
        resolved = set()
        for pkg in self.returnPackages():
            for name, flag, evr in pkg.requires:
                if not name.startswith('/'):
                    continue
                installedFileRequires.setdefault(pkg.pkgtup, []).append(name)
                if name not in resolved:
                    dep = self.getProvides(name, flag, evr)
                    resolved.add(name)
                    if not dep:
                        installedUnresolvedFileRequires.add(name)

        fileRequires = set()
        for fnames in installedFileRequires.itervalues():
            fileRequires.update(fnames)
        installedFileProviders = {}
        for fname in fileRequires:
            pkgtups = [pkg.pkgtup for pkg in self.getProvides(fname)]
            installedFileProviders[fname] = pkgtups

        ret =  (installedFileRequires, installedUnresolvedFileRequires,
                installedFileProviders)

        return ret

    def returnConflictPackages(self):
        '''Another method that RPMDB has but MetaSack doesn't.'''
        if self._cached_conflicts_data is None:
            ret = []
            for pkg in self.returnPackages():
                if len(pkg.conflicts):
                    ret.append(pkg)
            self._cached_conflicts_data = ret
        return self._cached_conflicts_data

    def transactionCacheFileRequires(self, installedFileRequires,
                                     installedUnresolvedFileRequires,
                                     installedFileProvides,
                                     problems):
        '''No-op - depcheck doesn't cache RPMDB data'''
        return
    def transactionCacheConflictPackages(self, pkgs):
        '''No-op - depcheck doesn't cache RPMDB data'''
        return
    def transactionReset(self):
        '''No-op - all this does is clear the cache, and depcheck doesn't
        cache data, so it doesn't need to do anything.'''
        return

def do_depcheck(repo, package_files, accepted=[], profile=False, test_dir=None,
                accepted_dir=None, **kwargs):
    '''Given a repoid and a list of package files, check the dependencies
    of the listed packages with respect to the given repo (and its parents).
    Args:
        repo: id of the repo to test against (see repoinfo.repos())
        package_files: a list of RPM files to test
    Optional args:
        accepted: a list of package filenames which are to be treated as if
          they are in already part of the repo(s) we're testing against
        profile: if True, run using yum's profiler, and emit profiling output
    Returns a dict with the following keys:
        retval: integer return value of YumBase.resolveDeps()
        problems: problem list from YumBase.resolveDeps()
        accept: list of package objects which passed the depcheck
        reject: list of package objects which failed depcheck
        ignored: list of package objects which were ignored by yum -
                 usually due to the package being older than what's in the
                 live repos
        prev_accepted: list of package objects for packages used as 'accepted'
                       (see above)
    '''
    print "Checking packages against repo %s (parents: %s)" % (repo,
              " ".join(repoinfo.getparents(repo)))
    yum_repos = set_up_yum_object(repo)

    # choose the correct arch(es) for the upcoming mashes
    if rpmarch.isMultiLibArch(yum_repos.arch.basearch):
        mash_arches = ' '.join(both_arches)
    else:
        mash_arches = yum_repos.arch.basearch

    # Add the accepted packages (if any) to the YumBase object
    prev_accepted = []
    if accepted:
        # create a tempdir if one has not been already passed in
        if accepted_dir is None:
            accepted_dir = tempfile.mkdtemp(prefix='depcheck-accepted.')

        # mash the accepted packages into a proto-updates repo
        for p in accepted:
            os.symlink(os.path.realpath(p), os.path.join(accepted_dir, os.path.basename(p)))
        do_mash(accepted_dir, mash_arches)
        accrepo = yum_repos.add_enable_repo('prev_accepted', ['file://%s' % accepted_dir])
        yum_repos.pkgSack # initializes package sacks
        prev_accepted = list(accrepo.sack)

    # This YumBase object will act like all repo packages are already installed
    y = YumDepcheck()
    y.tsInfo = yum.transactioninfo.TransactionData()
    y.tsInfo.debug = 1

    # yum_repos.pkgSack contains _all_ packages from repos 'repo' + parents
    # and also 'previously accepted' packages.
    # By moving the pkgSack into y.rpmdb, we're basically tricking yum into
    # thinking that all these packages are installed.
    #
    # This has one little problem - for almost every package, there are multiple
    # versions (sorted alphabetically in the pkgSack (= from oldest to latest)
    # while yum is searching for deps, it takes the first package found in
    # rpmdb (i.e. the oldest), which causes some weird errors.
    # To overcome this problem, we prefilter the pkgSack before saving it into
    # the rpmdb, so it only contains the latest version of each package.
    _rpmdb =  yum_repos.pkgSack

    # be aware that this method alters 'both' _rpmdb, and yum_repos.pkgSack
    # since they are both just pointer to the same instance in memory
    prefilter_rpmdb(_rpmdb)

    y.rpmdb = MetaSackRPMDB(_rpmdb)
    y.xsack = yum.packageSack.PackageSack()
    # Hacky way to set up the databases (this is copied from yum's testbase.py)
    y.tsInfo.setDatabases(y.rpmdb, y.xsack)

    # Set up some nice verbose output
    y.doLoggingSetup(9, 9)
    y.setupProgressCallbacks()
    # TODO: filter log messages down to the ones we actually care about

    # Mash the given packages into the set that would get pushed
    # TODO: try to handle debuginfo / src rpms? what would we need to check?

    # Set up a temp dir we can mash in if one hasn't been passed in
    # (so we don't delete our input files)
    if test_dir is None:
        test_dir = tempfile.mkdtemp(prefix='depcheck.')
    for p in package_files:
        os.symlink(os.path.realpath(p), os.path.join(test_dir, os.path.basename(p)))
    # mash away, you can mash away.. stay all day.. if you want to
    do_mash(test_dir, mash_arches)
    # Attempt to apply all the pending packages as updates
    pending = y.add_enable_repo('pending', ['file://%s' % test_dir])
    y.pkgSack # populates all package sacks
    ignored = []
    for p in pending.sack:
        if not y.update(p):
            if not y.install(p):
                ignored.append(p)

    # ENGAGE THRUSTERS WE ARE GO FOR LIFTOFF!! MOVE ZIG!! AND SO ON!!
    if profile:
        (r, problems) = y.cprof_resolveDeps()
    else:
        (r, problems) = y.resolveDeps()

    # C&P from YumBase.buildTransaction; possibly unnecessary (but doesn't hurt)
    y.rpmdb.ts = None
    # This is the skip-broken step
    if r == 1:
        y.skipped_packages = []
        (r, problems) = y._skipPackagesWithProblems(r, problems)
    # C&P from buildTransaction again: if we skipped broken packages, re-resolve
    if y.tsInfo.changed:
        (r, problems) = y.resolveDeps(r == 1)

    ok_states = ('u','i') # o, ud, od?
    ok_packages = [t.po for t in y.tsInfo.getMembers() if t.ts_state in ok_states]

    return dict(retval=r,
                problems=problems,
                prev_accepted=prev_accepted,
                ignored=ignored,
                reject=y.skipped_packages,
                accept=ok_packages)

def print_depcheck_results(accept=[], reject=[], ignored=[], prev_accepted=[], **kwargs):
    '''Print the results of do_depcheck in a nice easy(ish)-to-parse way'''
    if prev_accepted:
        print "PREVIOUSLY-ACCEPTED: %s" % " ".join([str(p) for p in prev_accepted])
    if ignored:
        print "IGNORE: %s" % " ".join([str(p) for p in ignored])
    print "REJECT: %s" % " ".join([str(p) for p in reject])
    print "ACCEPT: %s" % " ".join([str(p) for p in accept])

def depcheck_main(repo, package_files, accepted=[], profile=False):
    '''Convenience method to run the normal depcheck on the given package
    files, print the results, and return a tuple (retval, problems, accept)'''

    # create temp dirs for depcheck repositories
    temp_dir = tempfile.mkdtemp(prefix='depcheck.')
    acc_dir = tempfile.mkdtemp(prefix='depcheck-accepted.')
    print "Temp dirs %s and %s created." % (temp_dir, acc_dir)

    try:
        results = do_depcheck(repo, package_files, accepted, profile,
                                    test_dir=temp_dir, accepted_dir=acc_dir)
        print_depcheck_results(**results)

    finally:
        # remove the temp dirs created for depcheck
        print "removing temp dirs %s and %s" % (temp_dir, acc_dir)
        os.system('/bin/rm -rf %s' % temp_dir)
        os.system('/bin/rm -rf %s' % acc_dir)

    # Sorta dumb return values, retained for backwards compatibility
    return (results['retval'], results['problems'], results['accept']+results['prev_accepted'])
