#!/usr/bin/python

import os
import sys
import re
from rpmUtils import miscutils as rpm_misc
from copy import deepcopy
from tempfile import mkstemp
from autoqa.util import envra


class DepcheckPretty(object):
    def __init__(self, depcheck_output, pending_builds):
        """
        @param depcheck_output - string containing output of depcheck run
        @param pending_builds - dictionary;
          key = name of the build,
          value = list of rpm files in the build.
          Example:
            {'orc-0.4.14-1.fc14' : ['orc-devel-0.4.14-1.fc14.i686.rpm', 'orc-0.4.14-1.fc14.i686.rpm']}
        """

        f, fname = mkstemp()
        f = open(fname, 'w')
        f.write(depcheck_output)
        f.close()

        self.depcheck_output = depcheck_output
        self.pending_builds = pending_builds
        self.parsed = None
        self.digest = None

        self.parsed = self._parse_log(fname, pending_builds)
        self.digest = self._run_digest(self.parsed)

        os.remove(fname)

    def _parse_log(self, fname, pending_builds):
        """
        @param fname path to file containing depcheck output
        @param pending_builds - dictionary;
          key = name of the build,
          value = list of rpm files in the build.
          Example:
            {'orc-0.4.14-1.fc14' : ['orc-devel-0.4.14-1.fc14.i686.rpm', 'orc-0.4.14-1.fc14.i686.rpm']}
        @returns - parsed output - dictionary
          key = name of the build
          value = dictionary of information related to the build
            {'interest': [string, ...] (lines related to the build),
             'depsolving': [[string, ...], ...] (blocks of 'XYZ has depsolving problems'),
             'rpms': [string, ...] (rpm files related to the build),
             're': [compiled_regexp, ...] (regexps for 'interesting' lines),
            }
        """

        log = open(fname, 'r')

        line = True # do not change!

        # create dict with pending builds as keys, None as values {"foo":None, "bar":None}
        p_builds = dict().fromkeys(pending_builds.keys())

        # for each builds, add dictionary storing interesting data
        for build in p_builds:
            p_builds[build] = {'interest': [],
                               'depsolving': [],
                               'rpms': [], 
                               're': [
                                       re.compile(re.escape(build)),
                                       re.compile('TSINFO: Current Transaction'),
                                       re.compile('Dependency Process ending'),
                                       re.compile('Running transaction check'),
                                       re.compile('Packages skipped because of dependency problems'),
                                     ],
                              }
            for rpm in pending_builds[build]:
                p_builds[build]['re'].append(re.compile(re.escape(envra(rpm, 'n'))))


       # skip non-interesting part
        re_pkgsack = re.compile(r'Setting up Package Sacks')
        while line:
            line = log.readline()
            if re_pkgsack.search(line) is not None:
                break

        re_depsolving_problems = re.compile("has depsolving problems")
        re_arrow = re.compile(re.escape("-->"))

        # get the rest of lines containing build or rpm name
        while line:
            line = log.readline()
            # First of all, we need to check for the
            # 'XYZ has depsolving problems' line
            # If this line is found, this line, and all subsequend lines
            # starting with '-->' are stored in a buffer (log_lines).
            #
            # Because 'depsolving problems' can contain rpms from
            # multiple builds and/or other rpms, it's crucial
            # to scan all the lines in the 'depsolving problem'
            # buffer, and append the buffer to each affected pending build's
            # interesting lines.
            #
            # Because the inner while loop reads 'one more line' (i.e.
            # the first line not starting with '-->'). This line need's
            # to be completely re-examined, beacause it can contain
            # not only 'other information', but also next 'depsolving problems'
            # block. 
            # To achieve this effect, it's required to seek back in the file
            # and enforce 'jump' to the beginning of the outer loop. Hence the
            # try-catch block and `raise StopIteration` line.
            try:
                # Start of 'depsolving problems' block found
                if re_depsolving_problems.search(line) is not None:
                    # store the first line to buffer
                    log_lines = [line]
                    affected_builds = []

                    # as long, as the subsequent lines start with '-->'
                    # add them to the buffer.
                    line = log.readline()
                    while re_arrow.search(line) is not None:
                        log_lines.append(line)
                        line = log.readline()

                    # Seek back to the beginning of the lastly read line.
                    # This line is not a part of the 'depsolving problems'
                    # block, and needs to be completely re-examined
                    log.seek(log.tell() - len(line))

                    # determine which builds are affected by this depsolving
                    # problems block ...
                    for l in log_lines:
                        for build in p_builds:
                            for regexp in p_builds[build]['re']:
                                if regexp.search(l) is not None:
                                    affected_builds.append(build)
                                    break

                    # ... and store it to the interesting lines
                    # of each affected build
                    affected_builds = set(affected_builds)
                    for build in affected_builds:
                        p_builds[build]['depsolving'].append(log_lines)
                        p_builds[build]['interest'].extend(log_lines)

                    raise StopIteration()

                # simply check whether the line contains
                # build or rpm-from-the-build name
                for build in p_builds:
                    for regexp in p_builds[build]['re']:
                        if regexp.search(line) is not None:
                            p_builds[build]['interest'].append(line)
                            break

            except StopIteration:
                pass


        log.close()

        return p_builds

    def _run_digest(self, parsed_output):
        """
        @param parsed_output - dictionary returned from _parse_log()
        @returns dictionary
          key = name of the build
          value = dictionary
            {'highlights': list of strings ('depsolving problems' blocks),
             'all': string (interesting lines, some noise filtered out),
            }
        """
        p_builds = parsed_output

        digest = dict().fromkeys(p_builds.keys())

        re_skip_looking_for = re.compile("looking for.*as a requirement of.*")
        re_skip_already_ts = re.compile("Package.*already in ts")
        re_skip_already_updated = re.compile("Not Updating Package that is already updated")
        #fits lines starting with either ACCEPT, REJECT or IGNORE
        re_skip_sum_line = re.compile("^ACCEPT|^REJECT|^IGNORE|^PREVIOUSLY-ACCEPTED")


        re_tsinfo = re.compile("TSINFO: Current Transaction")
        re_dependency_ending = re.compile('Dependency Process ending')
        re_running_transaction = re.compile('Running transaction check')
        re_depsolving_problems = re.compile("has depsolving problems")


        re_skipped_packages_msg = re.compile("Packages skipped because of dependency problems:")


        for build in p_builds:
            digest[build] = {'highlights': [], 'all': ""}

            for depsolving_block in p_builds[build]['depsolving']:
                dep_blck = "%s" % "".join(depsolving_block)
                digest[build]['highlights'].append(dep_blck.rstrip())

            aux_len = len(p_builds[build]['interest'])

            for i in range(aux_len):
                line = p_builds[build]['interest'][i]
                if (
                        (re_skip_looking_for.search(line) is not None) or
                        (re_skip_already_ts.search(line) is not None) or
                        (re_skip_already_updated.search(line) is not None) or
                        (re_skip_sum_line.search(line) is not None)
                   ):
                    continue


                if (
                        (re_skipped_packages_msg.search(line) is not None) and
                        #list boundaries checking
                        ((i+1) < aux_len) and
                        (re_skip_sum_line.search(p_builds[build]['interest'][i+1]) is not None)
                   ):
                    continue


                if (
                        (re_depsolving_problems.search(line) is not None) or
                        (re_tsinfo.search(line) is not None) or
                        (re_running_transaction.search(line) is not None)
                   ):
                    digest[build]['all'] += "\n"


                digest[build]['all'] += line

                if (re_dependency_ending.search(line) is not None):
                    digest[build]['all'] += "\n"

        return digest


