#!/usr/bin/python
# depcheck_unittests.py - unittests for depcheck
#
# Copyright 2010, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Will Woods <wwoods@redhat.com>

# See depcheck_lib.py for all kinds of details and commentary.

import os
import sys
import rpmUtils.arch as rpmarch
import unittest
from autoqa.repoinfo import repoinfo
from depcheck_lib import depcheck_main, do_depcheck
from depcheck_lib import do_mash, both_arches, set_up_yum_object
import rpmfluff
import autoqa.util

# FIXME: This unittest works only for 64bit operating system arch (don't get
# confused with hardware architecture) - because of multilib tests. We need to
# detect operating system architecture and print out error if it is only 32bit.
# Or we could adjust those multilib tests to expect different results for that
# arch.

# === Unit tests begin here! Whee!
# --- unittest helper functions

def fake_package_update(po):
    '''Given a package object, use rpmfluff to generate a plausible
    update to that package. Returns a SimpleRpmBuild.'''
    # TODO: args to filter req/prov/etc.
    u = rpmfluff.SimpleRpmBuild(po.name, po.version, po.release+'.fakeupdate')
    u.buildArchs = [po.arch]
    for p in po.provides_print:
        u.add_provides(p)
    for r in po.requires_print:
        u.add_requires(r)
    for c in po.conflicts_print:
        u.add_conflicts(c)
    for o in po.obsoletes_print:
        u.add_obsoletes(o)
    for f in po.filelist:
        u.add_installed_file(f, rpmfluff.SourceFile(os.path.basename(f),''))
    for d in po.dirlist:
        u.add_installed_directory(d)
    u.make()
    try:
        rpm = u.get_built_rpm(po.arch)
        sn = os.path.basename(rpm)
        os.rename(rpm, sn)
    finally:
        u.clean()
    return sn

def add_local_repo(repoid, path, parents=[]):
    '''Add a fake entry to repoinfo for a local repo'''
    repoinfo.config.add_section(repoid)
    repoinfo.config.set(repoid, 'name', repoid)
    repoinfo.config.set(repoid, 'tag', 'fake-tag-%s' % repoid)
    repoinfo.config.set(repoid, 'url', 'file://%s' % path)
    repoinfo.config.set(repoid, 'path', '')
    repoinfo.config.set(repoid, 'collection_name', repoid.upper())
    repoinfo.config.set(repoid, 'parents', ", ".join(parents))

def simple_rpmlist():
    '''Create a set of simple RPM objects for testing'''
    m = rpmfluff.SimpleRpmBuild('mint', '1.0', '1')
    b = rpmfluff.SimpleRpmBuild('bourbon', '1.0', '1')
    j = rpmfluff.SimpleRpmBuild('julep', '1.0', '1')
    j.add_requires('bourbon = 1.0')
    j.add_requires('mint')
    return [m, b, j]

def multilib_rpmlist(arches):
    '''Create a multilib set of RPM objects'''
    g = rpmfluff.SimpleRpmBuild('gin', '1.0', '1')
    v = rpmfluff.SimpleRpmBuild('vermouth', '1.0', '1')
    v.add_subpackage('dry')
    v.add_subpackage('sweet')
    mt = rpmfluff.SimpleRpmBuild('martini', '1.0', '1')
    mt.add_requires('vermouth-dry = 1.0')
    mt.add_requires('gin')
    mt.section_install += 'mkdir -p "$RPM_BUILD_ROOT%{_libdir}"\n'
    mt.section_install += 'echo "Fake $RPM_ARCH library" > $RPM_BUILD_ROOT/%{_libdir}/libmartini.so.1\n'
    mt.basePackage.section_files += "%{_libdir}/libmartini.so.1\n"
    # TODO multilib binary (valid duplicate file)
    # multilib-ify
    mt.buildArchs = arches
    g.buildArchs = arches
    v.buildArchs = arches
    # add -devel subpackages, which mash handles specially
    g.add_devel_subpackage() # proper -devel subpackage which requires the main one
    v.add_subpackage('devel') # omits normal Requires: parent (for testing)
    return [g, v, mt]

class YumRepoBuildMultilib(rpmfluff.YumRepoBuild):
    def make(self, archlist):
        assert type(archlist) in (list, tuple, set), \
            "YumRepoBuildMultilib.make() takes a list argument"
        for pkg in self.rpmBuilds:
            pkg.make()
        for pkg in self.rpmBuilds:
            for arch in archlist:
                if (pkg.buildArchs is None and arch == rpmfluff.expectedArch) or \
                   (pkg.buildArchs and arch in pkg.buildArchs):
                    for n in pkg.get_subpackage_names():
                        rpmfluff.run_command('cp %s %s' % \
                          (pkg.get_built_rpm(arch, name=n), self.repoDir))
        return do_mash(self.repoDir, " ".join(archlist))

# --- Actual unittest testcases

class DepcheckPrereqTestCase(unittest.TestCase):
    '''A TestCase for the prerequisites for depcheck'''
    def test_rpmfluff(self):
        (n, v, r) = ('test-rpmfluff', '1.0', '1')
        p = rpmfluff.SimpleRpmBuild(n, v, r)
        p.clean()
        p.make()
        rpmsDir = p.get_rpms_dir()
        self.assertTrue(os.path.isdir(rpmsDir))
        arch = rpmfluff.expectedArch
        rpmFile = os.path.join(rpmsDir, arch, "%s-%s-%s.%s.rpm"%(n, v, r, arch))
        print rpmFile
        self.assertTrue(os.path.isfile(rpmFile))
        h = rpmfluff.get_rpm_header(rpmFile)
        self.assertEqual(h['name'], n)
        self.assertEqual(h['version'], v)
        self.assertEqual(h['release'], r)
        # clean up
        p.clean()

    def test_simple_yum_setup(self):
        arch = rpmfluff.expectedArch
        rpmlist = simple_rpmlist()
        repo = YumRepoBuildMultilib(rpmlist)
        repo.make([arch])
        # falsify repoinfo
        add_local_repo('simple_yum_setup', repo.repoDir)
        y = set_up_yum_object('simple_yum_setup')
        # Now see if we can get the RPM data back out through yum
        for r in rpmlist:
            plist = y.pkgSack.returnNewestByNameArch((r.name, arch))
            self.assertTrue(plist)
            for p in plist:
                self.assertEqual(p.name, r.name)
                self.assertEqual(p.version, r.version)
                self.assertEqual(p.release, r.release)
                if p.name == 'julep':
                    requires = p.returnPrco('requires')
                    self.assertTrue(('mint', None, (None, None, None)) in requires)
                    self.assertTrue(('bourbon', 'EQ', ('0', '1.0', None)) in requires)
        # cleanup
        repoinfo.config.remove_section('simple_yum_setup')
        os.system('rm -rf %s' % repo.repoDir)

    def test_mash(self):
        '''Check to see if mash/multilib works as we expect it to'''
        arch = rpmfluff.expectedArch
        is_multilib = rpmarch.isMultiLibArch(arch)
        rpmlist = multilib_rpmlist(both_arches)
        repo = YumRepoBuildMultilib(rpmlist)
        if is_multilib:
            rc = repo.make(both_arches)
        else:
            rc = repo.make([arch])
        self.assertEqual(rc, 0)
        add_local_repo('test_mash', repo.repoDir)
        y = set_up_yum_object('test_mash')
        # We should now have a proper multilib repo.
        # Testing this is tricky, since we're checking mash's notoriously
        # hard-to-understand behavior - but we'll check a few simple
        # things and assume it's doing what we want.
        # (first, prepare the data we're going to test)
        pkgarch = {}
        for p in y.pkgSack.returnNewestByNameArch():
            if p.name not in pkgarch:
                pkgarch[p.name] = set()
            pkgarch[p.name].add(p.arch)

        mainarch = set([arch])
        multilib = set(both_arches)
        # 1) should have both arches for -devel packages and packages with libs
        if is_multilib:
            self.assertEqual(pkgarch['martini'], multilib)
            self.assertEqual(pkgarch['gin-devel'], multilib)
            self.assertEqual(pkgarch['vermouth-devel'], multilib)
        else:
            self.assertEqual(pkgarch['martini'], mainarch)
            self.assertEqual(pkgarch['gin-devel'], mainarch)
            self.assertEqual(pkgarch['vermouth-devel'], mainarch)
        # 3) all other packages should be main-arch only
        self.assertEqual(pkgarch['gin'], mainarch)
        self.assertEqual(pkgarch['vermouth'], mainarch)
        self.assertEqual(pkgarch['vermouth-dry'], mainarch)
        self.assertEqual(pkgarch['vermouth-sweet'], mainarch)

        # clean up
        for r in rpmlist:
            r.clean()
        # no repo.clean, unfortunately
        repoinfo.config.remove_section('test_mash')
        os.system('rm -rf %s' % repo.repoDir)

class DepcheckTestCase(unittest.TestCase):
    def setUp(self):
        self.arch = rpmfluff.expectedArch
        self.rpmlist = simple_rpmlist()
        self.mainrepo = YumRepoBuildMultilib(self.rpmlist)
        self.mainrepo.make([self.arch])
        self.mainrepo.id = 'mainrepo'
        add_local_repo(self.mainrepo.id, self.mainrepo.repoDir)

    def tearDown(self):
        for rpm in self.rpmlist:
            rpm.clean()
        for repo in (self.mainrepo, ):
            os.system('rm -rf %s' % repo.repoDir) # no repo.clean()
            repoinfo.config.remove_section(repo.id)

    def test_depcheck_empty_transaction(self):
        '''Make sure we accept a consistent repo, and no updates at all'''
        (rv, problems, ok_packages) = depcheck_main(self.mainrepo.id, [])
        self.assertEqual(len(problems), 1)
        self.assertTrue(problems[0].startswith(u'Success'))# XXX subject to LANG?
        self.assertEqual(rv, 0)

    def test_depcheck_good_update(self):
        '''Make sure depcheck accepts a good update as expected'''
        (n, v, r) = ('mint', '2.0', '1') # Now twice as minty as mint 1.0!
        p = rpmfluff.SimpleRpmBuild(n, v, r)
        p.make()
        (rv, problems, ok_packages) = depcheck_main(self.mainrepo.id, [p.get_built_rpm(self.arch)])
        self.assertEqual(len(problems), 1)
        self.assertTrue(problems[0].startswith(u'Success'))# XXX subject to LANG?
        self.assertEqual(len(ok_packages), 1)
        okp = ok_packages[0]
        self.assertEqual((n, v, r),(okp.name, okp.version, okp.release))
        self.assertEqual(rv, 2)
        p.clean()

    def test_missing_req(self):
        '''make sure depcheck catches updates with missing requires'''
        (n, v, r) = ('mint', '3.0', '1') # Muddled for extra flavor!
        p = rpmfluff.SimpleRpmBuild(n, v, r)
        p.add_requires('muddler') # uh oh, this doesn't exist..
        p.make()
        (rv, problems, ok_packages) = depcheck_main(self.mainrepo.id, [p.get_built_rpm(self.arch)])
        self.assertEqual(len(ok_packages), 0)
        self.assertEqual(rv, 0)
        p.clean()

    def test_depcheck_new_dep(self):
        '''make sure depcheck accepts a good update with a new (good) dep'''
        muddler = rpmfluff.SimpleRpmBuild('muddler', '1.0', '1') # Yay, a muddler
        muddler.make()
        mint = rpmfluff.SimpleRpmBuild('mint', '3.0', '1') # Muddled for extra flavor!
        mint.add_requires('muddler') # NOW it exists
        mint.make()
        pkgs = [p.get_built_rpm(self.arch) for p in (muddler, mint)]
        (rv, problems, ok_packages) = depcheck_main(self.mainrepo.id, pkgs)
        for p in ok_packages: print p
        self.assertEqual(len(ok_packages), 2)
        self.assertEqual(rv, 2)
        mint.clean()
        muddler.clean()

    def test_changed_prov(self):
        '''Test a changed Provides:'''
        (n, v, r) = ('bourbon', '2.0', '1') # Browner than ever!
        p = rpmfluff.SimpleRpmBuild(n, v, r)
        p.make()
        (rv, problems, ok_packages) = depcheck_main(self.mainrepo.id, [p.get_built_rpm(self.arch)])
        self.assertEqual(len(ok_packages), 0)
        self.assertEqual(rv, 0)
        p.clean()

    def test_multiple_packages(self):
        '''Test with one good and one bad update'''
        (n, v, r) = ('bourbon', '2.0', '1') # Browner than ever!
        b = rpmfluff.SimpleRpmBuild(n, v, r)
        b.make()
        (n, v, r) = ('mint', '2.0', '1') # Now twice as minty as mint 1.0!
        m = rpmfluff.SimpleRpmBuild(n, v, r)
        m.make()
        (rv, problems, ok_packages) = depcheck_main(self.mainrepo.id,
                                    [p.get_built_rpm(self.arch) for p in (m,b)])
        self.assertEqual(len(ok_packages), 1)
        okp = ok_packages[0]
        self.assertEqual((n, v, r), (okp.name, okp.version, okp.release))
        m.clean()
        b.clean()

    def test_valid_package_conflict(self):
        '''Check to make sure we accept a valid package conflict'''
        b = rpmfluff.SimpleRpmBuild('bourbon', '3.0', '1')
        b.make()
        j = rpmfluff.SimpleRpmBuild('julep', '3.0', '1')
        j.add_requires('mint') # same as above
        j.add_conflicts('bourbon < 3.0') # this is new!
        j.make()
        (rv, problems, ok_packages) = depcheck_main(self.mainrepo.id,
                                    [p.get_built_rpm(self.arch) for p in (j,b)])
        self.assertEqual(len(ok_packages), 2)
        b.clean()
        j.clean()

    def test_bad_package_conflict(self):
        '''Check to make sure we reject an invalid package conflict'''
        j = rpmfluff.SimpleRpmBuild('julep', '4.0', '1')
        j.add_conflicts('mint < 3.0') # The only mint we have is 1.0!
        j.make()
        (rv, problems, ok_packages) = depcheck_main(self.mainrepo.id,
                                    [j.get_built_rpm(self.arch)])
        self.assertEqual(len(ok_packages), 0)
        j.clean()

    def test_bad_obsoletes(self):
        '''Make sure we reject Obsoletes that cause unresolved deps'''
        km = rpmfluff.SimpleRpmBuild('kentucky-colonel-mint', '1.0', '1')
        km.add_obsoletes('mint <= 5.0')
        km.make()
        (rv, problems, ok_packages) = depcheck_main(self.mainrepo.id,
                                    [km.get_built_rpm(self.arch)])
        km.clean()
        self.assertEqual(len(ok_packages), 0)

    def test_good_obsoletes(self):
        '''Make sure we accept good Obsoletes'''
        km = rpmfluff.SimpleRpmBuild('kentucky-colonel-mint', '2.0', '1')
        km.add_obsoletes('mint <= 5.0')
        km.add_provides('mint')
        km.make()
        (rv, problems, ok_packages) = depcheck_main(self.mainrepo.id,
                                    [km.get_built_rpm(self.arch)])
        km.clean()
        self.assertEqual(len(ok_packages), 1)

    def test_ignored(self):
        '''Check that ignored packages show up in ignored'''
        m = rpmfluff.SimpleRpmBuild('mint', '0.3', '0.alpha6.lol.wut') # old and moldy
        m.make()
        r = do_depcheck(self.mainrepo.id, [m.get_built_rpm(self.arch)])
        self.assertEqual(len(r['ignored']), 1)
        self.assertEqual(r['ignored'][0].name, 'mint')
        self.assertEqual(len(r['accept']), 0)
        m.clean()

    def test_accepted(self):
        '''Check that the --accepted flag properly protects packages'''
        # updating the whole delicious julep stack
        m = rpmfluff.SimpleRpmBuild('mint', '5.0', '1')
        b = rpmfluff.SimpleRpmBuild('bourbon', '5.0', '1')
        j = rpmfluff.SimpleRpmBuild('julep', '5.0', '1')
        j.add_requires('bourbon = 5.0')
        j.add_requires('mint')
        # ..but now there's a new mint - except the new julep build hasn't
        # landed yet
        km = rpmfluff.SimpleRpmBuild('kentucky-colonel-mint', '1.0', '1')
        km.add_obsoletes('mint <= 5.0')
        plist = (m, b, j, km)
        for p in plist: p.make()
        (rv, problems, ok_packages) = depcheck_main(self.mainrepo.id,
                            package_files=[km.get_built_rpm(self.arch)],
                            accepted=[p.get_built_rpm(self.arch) for p in (m,b,j)])
        should_be_ok = set(('mint', 'bourbon', 'julep'))
        for p in plist: p.clean()
        self.assertEqual(should_be_ok, set([p.name for p in ok_packages]))

class DepcheckMultilibTestCase(unittest.TestCase):
    '''More complex test cases that involve multilib packages'''
    def setUp(self):
        '''Set up a multilib repo for use with test cases'''
        self.arch = rpmfluff.expectedArch
        self.arches = both_arches
        self.is_multilib = rpmarch.isMultiLibArch(self.arch)
        self.rpmlist = simple_rpmlist() + multilib_rpmlist(self.arches)
        self.mainrepo = YumRepoBuildMultilib(self.rpmlist)
        if self.is_multilib:
            self.mainrepo.make(self.arches)
        else:
            self.mainrepo.make([self.arch])
        self.mainrepo.id = 'mainrepo'
        add_local_repo(self.mainrepo.id, self.mainrepo.repoDir)

    def tearDown(self):
        for rpm in self.rpmlist:
            rpm.clean()
        for repo in (self.mainrepo, ):
            os.system('rm -rf %s' % repo.repoDir) # no repo.clean()
            repoinfo.config.remove_section(repo.id)

    def test_empty_transaction(self):
        '''Make sure we accept an empty update transaction'''
        (rv, problems, ok_packages) = depcheck_main(self.mainrepo.id, [])
        self.assertEqual(len(problems), 1)
        self.assertTrue(problems[0].startswith(u'Success'))# XXX subject to LANG?
        self.assertEqual(rv, 0)

    def test_multilib_ok(self):
        '''Test an acceptable update to a multilib package'''
        (n, v, r) = ('gin', '10.0', '1') # Switching to Tanqueray 10
        p = rpmfluff.SimpleRpmBuild(n, v, r)
        p.buildArchs = self.arches
        p.add_devel_subpackage()
        p.make()
        pkglist = [p.get_built_rpm(a, name=n) for n in p.get_subpackage_names() for a in self.arches]
        (rv, problems, ok_packages) = depcheck_main(self.mainrepo.id, pkglist)
        self.assertEqual(len(problems), 1)
        self.assertTrue(problems[0].startswith(u'Success'))# XXX subject to LANG?
        if self.is_multilib:
            # mainarch gin, multilib gin-devel
            self.assertEqual(len(ok_packages), 3)
        else:
            # single arch gin + gin-devel
            self.assertEqual(len(ok_packages), 2)
        for okp in ok_packages:
            self.assertEqual((v, r), (okp.version, okp.release))
            self.assertTrue(okp.name in ('gin', 'gin-devel'))
        self.assertEqual(rv, 2)
        p.clean()

    def test_multilib_dropped_arch(self):
        '''Test an update which drops one of its arches'''
        # NOTE: this test makes no sense on non-multilib arches, so skip it.
        # XXX TODO skip this test properly.
        # py2.7 unittest has a skip() decorator (also in unittest2 for py2.6)
        if not self.is_multilib:
            return
        # This tries to simulate the nss-softokn error - see bug #596840
        v = rpmfluff.SimpleRpmBuild('vermouth', '1.0', '2')
        v.buildArchs = self.arches[1:] # oops, dropped one
        # Note that the subpackages intentionally do *not* require the main
        # package - that was a key component of the nss-softokn error.
        v.add_subpackage('dry')
        v.add_subpackage('sweet')
        v.add_subpackage('devel')
        v.make()
        pkglist = [v.get_built_rpm(a, name=n) for n in v.get_subpackage_names() for a in v.buildArchs]
        (rv, problems, ok_packages) = depcheck_main(self.mainrepo.id, pkglist)
        # NOTE: we actually expect vermouth to *pass*, since an update of this
        # type is *technically valid* - it only becomes problematic when a
        # subsequent update requires the missing library.
        # This shows that depcheck alone will not be sufficient to prevent
        # this situation from occurring in the future.
        self.assertEqual(len(ok_packages), 4)
        for okp in ok_packages:
            self.assertTrue(okp.name.startswith('vermouth'))
        # Simulate the presence of the vermouth package in an update repo
        updaterepo = YumRepoBuildMultilib([v])
        updaterepo.make(self.arches)
        updaterepo.id = 'updaterepo'
        add_local_repo(updaterepo.id, updaterepo.repoDir, parents=(self.mainrepo.id,))
        # So now the 'martini' packages ("a subsequent update") should fail,
        # since martini.i386 requires the missing vermouth-dry.i386 library..
        mt = rpmfluff.SimpleRpmBuild('martini', '3.0', '1')
        mt.add_requires('vermouth-dry = 1.0-2')
        mt.add_requires('gin')
        mt.section_install += 'mkdir -p "$RPM_BUILD_ROOT%{_libdir}"\n'
        mt.section_install += 'echo "Fake $RPM_ARCH library" > $RPM_BUILD_ROOT/%{_libdir}/libmartini.so.3\n'
        mt.basePackage.section_files += "%{_libdir}/libmartini.so.3\n"
        mt.buildArchs = self.arches
        mt.make()
        pkglist = [mt.get_built_rpm(a) for a in mt.buildArchs]
        (rv, problems, ok_packages) = depcheck_main(updaterepo.id, pkglist)
        # NOTE WELL: we expect yum to *accept* martini-3.0, because it
        # doesn't handle dependency resolution the same way RPM does.
        # (The nss-softokn failure happened inside the RPM transaction test.)
        # Therefore this test case just proves that depcheck/yum alone is
        # not sufficient to catch all dependency problems.
        # We should use this same testcase with whatever future tool actually
        # does the RPM transaction test, and expect it to reject martini-3.0.
        self.assertEqual(len(ok_packages), len(mt.buildArchs))
        # clean up updaterepo
        os.system('rm -rf %s' % updaterepo.repoDir)
        repoinfo.config.remove_section(updaterepo.id)
        # and clean up our packages
        v.clean()
        mt.clean()

if __name__ == '__main__':
    # this test requires amd64 machine because of multilib tests
    if autoqa.util.get_basearch() != 'x86_64':
        sys.exit('Error: This test can run only on a x86_64 machine.')

    unittest.main()
