#!/usr/bin/python
#
# Copyright 2012, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Will Woods <wwoods@redhat.com>  Hongqing YANG <hoyang@redhat.com>
'''Rawhide Acceptance Test Suite - install testing
   basic usage: ./install.py treeurl'''

import os
import sys
import time
import optparse
import anaconda
import pickle
import subprocess
from autoqa import util
from autoqa.config import Config
from autoqa.virtguest import VirtGuest


def _get_parser():
    parser = optparse.OptionParser(usage="%prog [options] TREEURL")
    display = ['xvfb']
    archlist = ['i386', 'x86_64', 'ppc']
    parser.add_option("--name", action="store",
        default='AutoQA',
        help='guest name')
    parser.add_option("-a", "--arch", action="store", type="choice",
        choices=archlist, default=util.get_basearch(),
        help="target architecture (%default)")
    parser.add_option("-r", "--ram", action="store",
        default=1024,
        help="guest memory size")
    parser.add_option("-d", "--disksize", action="store",
        default=10,
        help="disk size")
    parser.add_option("-k", "--ksfile", action="store",
        default='',
        help="kickstart file for installation")
    parser.add_option("-s", "--scratchdir", action="store",
        default=Config.paths.tmpdir,
        help="scratch dir for downloaded/generated files (%default)")
    parser.add_option("-l", "--logdir", action="store",
        default=os.getcwd(),
        help="write logs to this dir (%default)")
    parser.add_option("-b", "--bindir", action="store",
        default=os.getcwd(),
        help="binary dir - contains ks.cfg, etc. (%default)")
    parser.add_option("--repo", action="store",
        default=None,
        help="repository path")
    parser.add_option("-t", action="store_false",
        default=True, dest="delete",
        help="Delete guest after installation completation")
    parser.add_option("-x", "--extra-args", action="store",
        default="",
        help="Additional arguments to pass to the kernel")
    parser.add_option("--display", action="store", type="choice", choices=display,
        dest="display",
        help="Use an alternate DISPLAY (default: %default)")
    parser.add_option("-u", "--updates", action="append",
        default=None,
        help="updates.img locations")
    parser.add_option("--serial", action="store_false",
        default=True,
        help="use serial console or not")
    return parser


def parse_args(args=None, ks=None):
    parser = _get_parser()
    (opt, args) = parser.parse_args(args)

    # Validate arguments
    if len(args) != 1:
        parser.error("incorrect number of arguments")

    # Process TREEURL
    baseurl = args[0]
    if not baseurl.endswith('/') and not baseurl.endswith('.iso'):
        baseurl = baseurl + '/'

    # detect arch from tree url or media name
    if baseurl:
        if baseurl.find('i386')!= -1:
            opt.jobarch = 'i386'
        elif baseurl.find('i686')!=-1:
            opt.jobarch = 'i686'
        elif baseurl.find('x86_64')!=-1:
            opt.jobarch = 'x86_64'
        else:
            print "cannot recognize the Fedora architecture from baseurl"

    anaconda.scratchdir = opt.scratchdir
    anaconda.logdir = opt.logdir

    # if ksfile is not specifited, use the local default one
    if not opt.ksfile and ks:
        opt.ksfile = os.path.join(opt.bindir, ks)

    opt.keyval_file = os.path.join(opt.logdir, 'keyval')

    return (opt, baseurl)


def create_guest(location, name, ram, disksize, kickstart, repo, updates,
                 extra_args, serial):
    '''
    Attempt to create a new virt guest using the given data

    @param location: A directory containing install images
    @param name: Preferred guest name.  If a guest with matching name already
            exists, a new name will be selected by appending a unique number
    @param ram: Size of memory (in megabytes) to allocate for guest
    @param disksize: Size of virtue disk
    @param kickstart: Path to kickstart file to use for install
    @param repo: Repo path for installation
    @param updates: Path to updates.img
    @param extra_args: Additional boot arguments for the guest
    @param serial: Boolean value to indicate using serial or not
    @return: Object of type autoqa.virtguest.VirtGuest
    '''
    # Write kickstart
    print "writing kickstart to log"
    anaconda.log_ksfile(kickstart)
    # Images all ready. Create the virt guest.
    counter = 0
    guest = VirtGuest(name=name)
    while guest.exists():
        print "  guest name %s in use" % guest.name
        counter += 1
        guest = VirtGuest(name="%s_%u" % (name, counter))
    print "  new guest name: %s" % guest.name
    print "starting virt guest"
    guest.create(location, kickstart, ram, disksize, repo, updates, extra_args,
                 serial)
    guest.steplist = [] # A place to save the list of anaconda steps
    return guest


def system_sanity():
    '''Check the system to be sure the required binaries/services are there'''

    print 'Checking whether system contains all required tools...',
    try:
        msg = "'virsh --version' failed. Be sure libvirt-client is installed.'"
        assert subprocess.call(['virsh', '--version'],
                               stdout=subprocess.PIPE) == 0

        msg = "'virsh list' failed. Be sure the libvirtd service is running.'"
        assert subprocess.call(['virsh', 'list'],
                               stdout=subprocess.PIPE) == 0

        msg = "'service rsyslog status' is not active. Be sure rsyslog is running."
        assert subprocess.call(['service', 'rsyslog', 'status'],
                               stdout=subprocess.PIPE) == 0

        print 'OK'
        return True
    except (IOError, OSError, AssertionError):
        print 'FAIL'
        print msg
        return False


#http://fedoraproject.org/wiki/QA:Rawhide_Acceptance_Test_Plan#Installability_2
def main(args=None, ks=None):

    # Process command-line arguments
    (opt, baseurl) = parse_args(args, ks)

    if not system_sanity():
        return False

    # configure rsyslog service
    print "Configuring rsyslog...",
    anaconda.kill_rsyslog_port()
    anaconda.configure_rsyslog()
    print "done"

    # Start testing
    keyvals = {}
    print "== installer image sanity test =="
    timestamp = util.test("images", anaconda.installer_image_test(opt.jobarch, baseurl),
                          keyvals)
    keyvals["treeinfo_timestamp"] = timestamp
    keyvals["tree_time"] = int(float(timestamp))
    keyvals["treeinfo_localtime"] = time.ctime(timestamp)

    print 'Creating virtual guest...'
    sys.stdout.flush()

    guest = create_guest(baseurl, opt.name, opt.ram, opt.disksize, opt.ksfile,
                         opt.repo, opt.updates, opt.extra_args, opt.serial)

    passed = False
    try:
        print "\n== kernel boot test =="
        util.test("kernel_boot", anaconda.kernel_boot_test(guest), keyvals)
        keyvals["guestname"] = guest.name
        keyvals["kernelver"] = guest.kernelver

        print "\n== anaconda loading stage2 test =="
        # XXX raise timeout if unaccel? Usually doesn't take >5min to probe
        # network dev / start download, even in qemu..
        util.test("load_stage2", anaconda.loader_stage2_test(guest), keyvals)

        print "\n== anaconda disk probe test =="
        # XXX raise timeout if unaccel - it can take ~5min to get to %pre!
        util.test("diskprobe", anaconda.anaconda_disk_probe(guest), keyvals)

        # XXX raise timeout for unaccel - can take 2-3hrs to finish
        print "\n== anaconda package installation test =="
        util.test("package_install", anaconda.anaconda_finish_package_install(guest),
                  keyvals)

        print "\n== anaconda bootloader setup test =="
        util.test("bootloader", anaconda.anaconda_bootloader_setup(guest), keyvals)

        # FIXME bootloader test should include waiting for initrd after reboot
        print "\n== completing installation =="
        util.test("install_complete", anaconda.anaconda_install_finish(guest),
                  keyvals)

        passed = True
    except KeyboardInterrupt:
        print "\nTest is interrupted. Deleting failed guest."
        guest.delete()
        raise
    except Exception:
        print "\nCaught exception. Deleting failed guest."
        guest.delete()
        raise
    finally:
        # save keyvals
        keyvals.update(anaconda.keyvals)
        keyval_file = open(opt.keyval_file, 'w')
        pickle.dump(keyvals, keyval_file)
        keyval_file.close()
        # filter non-printable characters from logs
        for log in ('stage1.log', 'stage2.log'):
            log = os.path.join(opt.logdir, log)
            if os.path.exists(log):
                strings(log, log+'_filtered')
        # if we failed, include last lines from last log
        if not passed:
            print 'INSTALLATION FAILED'
            # find last log
            last_log = None
            for log in ('stage2.log_filtered', 'stage1.log_filtered', 'boot.log'):
                log = os.path.join(opt.logdir, log)
                if os.path.exists(log):
                    last_log = log
                    break
            # print last messages
            if last_log:
                print ('\nThese are the last lines from %s:' %
                        os.path.basename(last_log))
                log = open(last_log)
                lines = log.readlines()
                log.close()
                lines = lines[-10:] # last ten lines
                for line in lines:
                    print '# ' + line.strip()


    print "\n== testing complete =="
    print "Cleaning up guest."
    guest.delete()

    return True


def strings(in_file, out_file):
    '''Read input file, select only printable strings (using 'strings' command),
    filter out lines containing only ANSI control characters, and save it to
    output file.
    @param in_file file name to read text from
    @param out_file file name to save filtered text into
    '''
    # run 'strings'
    process = subprocess.Popen(['strings', in_file], stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT)
    (stdoutdata, stderrdata) = process.communicate()
    lines = stdoutdata.splitlines(True)

    # filter out ANSI escape sequences
    text = []
    for line in lines:
        if len(line.strip()) == 0:
            # just empty line
            continue
        if line.startswith('[') and len(line.strip()) <= 10:
            # probably just an escape sequence, filter out
            continue
        else:
            text.append(line)

    # save output
    file_ = open(out_file, 'w')
    file_.writelines(text)
    file_.close()


if __name__ == '__main__':
    success = main()
    sys.exit(0 if success else 1)
