#!/usr/bin/python
# Rawhide Acceptance Test Suite - repo sanity testing
#
# basic usage: ./sanity.py treeurl
#
# Copyright 2009, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Will Woods <wwoods@redhat.com>

import os
import yum
import optparse
import pickle
from autoqa import util
from autoqa.config import Config
import subprocess
from resolve_out_deps import resolve_deps # XXX this will be in yum someday

compsfile = None # dumb way to cache this between test functions
my = None # YumBase object
grabber = None
scratchdir = None
keyvals = {}

def _get_parser():
    parser = optparse.OptionParser(usage="%prog [options] TREEURL")
    archlist = ['i386', 'x86_64', 'ppc']
    parser.add_option("-a", "--arch", action="store", type="choice",
        choices=archlist, default=util.get_basearch(),
        help="target architecture (%default)")
    parser.add_option("-s", "--scratchdir", action="store",
        default=Config.paths.tmpdir,
        help="scratch dir for downloaded/generated files (%default)")
    parser.add_option("-l", "--logdir", action="store",
        default=os.getcwd(),
        help="write logs to this dir (%default)")
    return parser


def parse_args(args=None):
    parser = _get_parser()
    (opt, args) = parser.parse_args(args)

    # Validate arguments
    if len(args) != 1:
        parser.error("incorrect number of arguments")

    # Process TREEURL
    baseurl = args[0]
    if not baseurl.endswith('/') and not baseurl.endswith('.iso'):
        baseurl = baseurl + '/'

    global scratchdir
    scratchdir = opt.scratchdir

    opt.keyval_file = os.path.join(opt.logdir, 'keyval')

    return (opt, baseurl)


def repodata_validity_test():
    '''https://fedoraproject.org/wiki/QA:Repodata_validity_test_case'''
    import xml.etree.cElementTree as ElementTree
    data = {}
    ns = '{http://linux.duke.edu/metadata/repo}'
    repodata = grabber.urlopen('repodata/repomd.xml')
    print "fetching repodata"
    tree = ElementTree.parse(repodata)
    root = tree.getroot()
    print "parsing repomd.xml"
    revision = root.find(ns+'revision')
    print "repomd.xml timestamp: %s" % revision.text
    keyvals['repodata_time'] = revision.text
    for e in root.findall(ns+'data'):
        etype = e.attrib['type']
        if etype == 'primary_db' or (etype == 'primary' and etype not in data):
            data['primary'] = e
        if etype == 'other_db' or (etype == 'other' and etype not in data):
            data['other'] = e
        if etype == 'group_gz' or (etype == 'group' and etype not in data):
            data['group'] = e
        if etype == 'filelists_db' or (etype == 'filelists' and etype not in data):
            data['filelists'] = e
    for data_type in ('primary', 'other', 'group', 'filelists'):
        data_elem = data.get(data_type, None)
        util.check('<data type="%s"> exists' % data_type, data_elem is not None)
        loc = data_elem.find(ns+'location')
        util.check('%s has <location>' % data_type, loc is not None)
        util.check('<location> has href attribute', 'href' in loc.attrib)
        chcksum = data_elem.find(ns+'checksum')
        util.check('%s has <checksum>' % data_type, chcksum is not None)
        util.check('<checksum> has valid data', util.ishexstring(chcksum.text))
        print "fetching data"
        localfile = util.download(loc.attrib['href'], scratchdir, grabber=grabber)
        # dumb caching - but hey, it works
        if data_type == 'group':
            global compsfile
            compsfile = localfile
        algo = chcksum.attrib['type']
        if algo == 'sha':
            algo = 'sha1'
        util.checkhash(localfile, algo, chcksum.text)
    return True

def comps_validity_test():
    '''https://fedoraproject.org/wiki/QA:Comps_Validity_Test_Case'''
    # grab the relax NG schema
    rng_url = 'http://git.fedorahosted.org/git/?p=comps.git;a=blob_plain;'\
              'f=comps.rng;hb=HEAD'
    print "fetching RELAX NG schema for comps"
    rng_file = util.download(rng_url, os.path.join(scratchdir,'comps.rng'))
    global compsfile
    cmd = ['xmllint', '--noout', '--nowarning', '--relaxng',
           rng_file, compsfile]
    retval = subprocess.call(cmd)
    return util.check("comps syntax and grammar", retval == 0)

def critpath_closure_test(start_list):
    '''https://fedoraproject.org/wiki/QA:Core_package_dependency_closure_test_case'''
    name_list = []
    # Expand the start_list to a list of names
    for name in start_list:
        if name.startswith('@'):
            print "expanding %s" % name
            count = 0
            group = my.comps.return_group(name[1:])
            for groupmem in (group.mandatory_packages.keys() +
                             group.default_packages.keys()):
                if groupmem not in name_list:
                    name_list.append(groupmem)
                    count += 1
            print "%s packages added" % count
        else:
            if name not in name_list:
                name_list.append(name)
    # Iterate over the name_list
    count = 0
    pkg_list = []
    skipped_list = []
    for name in name_list:
        count += 1
        print "depsolving %4u/%4u (%s)" % (count, len(name_list), name)
        p = my.pkgSack.searchNevra(name=name)
        if not p:
            print "WARNING: unresolved package name: %s" % name
            skipped_list.append(name)
            continue
        for pkg in p:
            pkg_list.append(pkg)
            for dep in resolve_deps(pkg, my):
                if dep not in name_list:
                    print "    added %s" % dep
                    name_list.append(dep)

    print "depsolving complete."

    # Throw out the unresolved packages
    for name in skipped_list:
        name_list.remove(name)

    # TODO this isn't coming out right for i386
    print "%u packages in critical path (+%u multiarch)" % (len(name_list),
          len(pkg_list)-len(name_list))
    print "%u rejected package names: %s" % (len(skipped_list),
                                             " ".join(skipped_list))
    return pkg_list

def critpath_integrity_test(critpath_pkgs):
    '''https://fedoraproject.org/wiki/QA:Core_package_existence_test_case'''
    all_ok = True
    for pkg in critpath_pkgs:
        try:
            # check that the packages are present
            localfile = util.download(pkg.relativepath, scratchdir, grabber=grabber)
            # check checksum from metadata
            (algo, hexhash) = pkg.returnIdSum() # (algo, hexhash) tuple
            util.checkhash(localfile, algo, hexhash)
            # TODO: verify with RPM?
        except Exception, e:
            print "ERROR: %s" % str(e)
            all_ok = False
    return all_ok

if __name__ == '__main__':
    (opt, baseuri) = parse_args()

    my = yum.YumBase()
    if opt.arch != util.get_basearch():
        # try to force yum to use the supplied arch rather than the host arch
        fakearch = {'i386':'i686',  'x86_64':'x86_64',  'ppc':'ppc64'}
        my.preconf.arch = fakearch[opt.arch]
    my.conf.cachedir = scratchdir
    my.repos.disableRepo('*')
    my.add_enable_repo('rats-repo-%s' % opt.arch, baseurls=[baseuri])

    grabber = util.get_urlgrabber(prefix=baseuri)

    # https://fedoraproject.org/wiki/QA:Rawhide_Acceptance_Test_Plan#Repo_Sanity_2

    print "== repodata validity test =="
    util.test("repodata", repodata_validity_test(), keyvals)

    print "\n== comps validity test =="
    util.test("comps", comps_validity_test(), keyvals)

    (maj, minor, sub) = yum.__version_info__
    if (maj < 3 or minor < 2 or sub < 24) and opt.arch != util.get_basearch():
        print "\nWARNING: yum < 3.2.24 may be unable to depsolve other arches."
        print "The following tests will probably fail."
        print "Get a newer yum or run this on an actual %s system." % opt.arch

    print "\n== critical path dependency test =="
    critpath_pkgs = critpath_closure_test(['@core', '@critical-path-base',
                                           '@critical-path-gnome'])
    util.test("critpath_deps", critpath_pkgs, keyvals)
    namelist = set([p.name.encode('utf8') for p in critpath_pkgs])
    keyvals["critpath_count"] = len(namelist)

    logfile = open(os.path.join(opt.logdir, "critpath.log"), 'w')
    logfile.write("\n".join(sorted(namelist)))
    logfile.close()

    print "\n== critical path integrity test =="
    util.test("critpath_integrity", critpath_integrity_test(critpath_pkgs),
              keyvals)

    print "\n== testing complete =="

    # save keyvals
    keyval_file = open(opt.keyval_file, 'w')
    pickle.dump(keyvals, keyval_file)
    keyval_file.close()
