# -*- coding: utf-8 -*-
#
# Copyright 2009, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Kamil Paral <kparal@redhat.com>

from autotest_lib.client.bin import utils
import autoqa.koji_utils
import autoqa.util as util
import os
from autoqa.test import AutoQATest
from autoqa.decorators import ExceptionCatcher

class rpmguard(AutoQATest):

    @ExceptionCatcher()
    def setup(self, **kwargs):
        utils.system('yum -y install rpmlint')


    @ExceptionCatcher()
    def initialize(self, config, nvr, **kwargs):
        super(self.__class__, self).initialize(config)
        self.detail.id = nvr

        self.koji = autoqa.koji_utils.SimpleKojiClientSession()
        self.koji.check_connection()
        self.rpmguard = os.path.join(self.bindir, 'rpmguard')


    @ExceptionCatcher()
    def run_once(self, kojitag, nvr, name, **kwargs):
        super(self.__class__, self).run_once()

        # add header
        self.log('%s\n%s\n%s' % ('='*40, nvr, '='*40))

        # run the test for this nvr
        self.test_nvr(nvr, name, kojitag)

        # create self.summary
        if self.detail.result != 'ABORTED':
            # count the warnings
            output_lines = '\n'.join(self.detail.output).splitlines()
            warn_count = len([line for line in output_lines
                              if line.startswith('W: ')])
            self.detail.summary = '%d warnings' % warn_count

        self.detail.keyvals = {'kojitag': kojitag, 'envr': nvr}

        # submit results, including email opt-in
        self.post_results(optin={'pkg': name, 'kojitag': kojitag})


    def test_nvr(self, nvr, name, kojitag):
        '''Test a single NVR.

        @param nvr build NVR (string)
        @param name package name (string)
        @param kojitag Koji tag of this build (string)
        '''

        # temporary results for each architecture
        arch_results = {}

        # get the most recent release available
        evr = (util.envra(nvr,'e'), util.envra(nvr,'v'), util.envra(nvr, 'r'))
        lastBuild = self.koji.list_previous_release(name, kojitag,
            max_evr=(evr[0], evr[1], evr[2]))

        # if there is no such build, we don't have anything to compare
        if not lastBuild:
            self.log('N: There is no previous build of %s in %s tag (or its '
                     'parents).' % (nvr, kojitag))
            self.detail.update_result('PASSED')
            return

        # now we need list of RPMs available for each build
        new_rpms = self.koji.nvr_to_rpms(nvr, src=False)
        old_rpms = self.koji.nvr_to_rpms(lastBuild['nvr'], src=False)
        # and match the RPMs according to build name and architecture as
        # (old one, new one)
        rpm_to_match = []

        # traverse all RPMs in the newer build
        for new_rpm in new_rpms:
            # find the matching older RPM, there should be exactly one
            old_rpm = [r for r in old_rpms if r['arch'] == new_rpm['arch']
                        and r['name'] == new_rpm['name']]
            # we should mark all these packages as already matched
            for o in old_rpm:
                assert not o.has_key('seen'), ('The foreign object already has '
                                               'the same attribute as we use')
                o['seen'] = True
            # older RPM may not exist
            if not old_rpm:
                self.log('N: There is no previous RPM for %s.%s' %
                        (new_rpm['nvr'], new_rpm['arch']))
                continue
            # there certainly shouldn't be more than 1 matching older RPM, if there is,
            # we have a bug somewhere
            assert len(old_rpm) <= 1, ('There are %d existing older RPMs for '
                                       '%s.%s: %s' % (len(old_rpm),
                                       new_rpm['nvr'], new_rpm['arch'], old_rpm))

            # now there is certainly one older RPM matching the newer one
            old_rpm = old_rpm[0]
            rpm_to_match.append((old_rpm, new_rpm))

        # there may be some older RPMs not longer available in the newer build
        for old_rpm in old_rpms:
            if not old_rpm.has_key('seen'):
                self.log('N: There is no new RPM to match %s.%s' %
                         (old_rpm['nvr'], old_rpm['arch']))

        # let's compare the packages finally
        for old_rpm, new_rpm in rpm_to_match:

            # fetch old and new rpms
            old_file = util.download(old_rpm['url'], self.tmpdir)
            new_file = util.download(new_rpm['url'], self.tmpdir)

            # run rpmguard
            cmd = '%s %s %s' % (self.rpmguard, old_file, new_file)
            cmd_result = utils.run(cmd+" 2>&1", ignore_status=True,
                stdout_tee=utils.TEE_TO_LOGS, stderr_tee=utils.TEE_TO_LOGS)

            # store the results in temporary arch_results
            key = (old_rpm['nvr'], new_rpm['nvr'])
            if not arch_results.has_key(key):
                arch_results[key] = {}
            archs = arch_results[key]
            archs[old_rpm['arch']] = cmd_result.stdout

            # check command return code
            if cmd_result.exit_status == 0:
                self.detail.update_result('PASSED')
            elif cmd_result.exit_status == 10:
                self.detail.update_result('INFO')
            else:
                self.log('Unexpected exit status: %s' % cmd_result.exit_status,
                     stderr=True, highlight=True)
                self.detail.update_result('ABORTED')

        # Copy arch_results to results and eliminate duplicates
        for (old_nvr, new_nvr) in arch_results.keys():
            res_arch = {}
            for arch, res in arch_results[(old_nvr, new_nvr)].iteritems():
                if not res_arch.has_key(res):
                    res_arch[res] = [arch]
                else:
                    res_arch[res].append(arch)
            for res, archs in res_arch.iteritems():
                archlist = ''
                for arch in archs:
                    archlist += arch + ', '
                archlist = archlist[:-2]
                self.log('N: Comparing %s and %s (archs: %s) ...' %
                         (old_nvr, new_nvr, archlist), printout=False)
                if res:
                    # strip newline at the end
                    self.log(res.rstrip(), printout=False)
                self.log('N: ----', printout=False)

        return
